"""Modules for handing crystallographic lattice-parameters."""

import numpy as np
from numpy import dot
from collections import OrderedDict


def abs_cap(val, max_abs_val=1):
    """
    Return the value with its absolute value capped at max_abs_val.

    Particularly useful in passing values to trignometric functions where
    numerical errors may result in an argument > 1 being passed in.

    Args:

        val (float): Input value.

        max_abs_val (float): The maximum absolute value for val. Defaults to 1.

    Returns:
        val if abs(val) < 1 else sign of val * max_abs_val.
    """
    return max(min(val, max_abs_val), -max_abs_val)


class Lattice(object):
    """Construct Lattice parameter object."""

    def __init__(self, lattice_mat=None, round_off=5):
        """
        Hold lattice parameter information.

        >>> box=[[10,0,0],[0,10,0],[0,0,10]]
        >>> lat=Lattice(box)
        >>> lat.lat_lengths()
        [10.0, 10.0, 10.0]
        >>> Lattice(box).lattice()[0][0]
        10.0
        >>> lat.lat_angles()
        [90.0, 90.0, 90.0]
        >>> round(lat.inv_lattice()[0][0],2)
        0.1
        >>> [round(i,2) for i in lat.lat_angles(radians=True)]
        [1.57, 1.57, 1.57]
        >>> frac_coords=[[0,0,0],[0.5,0.5,0.5]]
        >>> lat.cart_coords(frac_coords)[1][1]
        5.0
        >>> cart_coords=[[0,0,0],[5,5,5]]
        >>> lat.frac_coords(cart_coords)[1][1]
        0.5
        >>> box=[[1e-8,0,0],[0,1e-8,0],[0,0,1e-8]]
        >>> lat=Lattice(box)
        >>> [round(i,2) for i in lat.lat_angles()]
        [90.0, 90.0, 90.0]
        """
        tmp = np.array(lattice_mat, dtype=np.float64).reshape((3, 3))
        self._lat = np.around(tmp, decimals=round_off)
        self._inv_lat = None
        self._lll_matrix_mappings = {}

    def lat_lengths(self):
        """Return lattice vectors' lengths."""
        return [
            round(i, 6)
            for i in (np.sqrt(np.sum(self._lat ** 2, axis=1)).tolist())
        ]
        # return [round(np.linalg.norm(v), 6) for v in self._lat]

    @property
    def volume(self):
        """Return volume given a lattice object."""
        m = self._lat
        vol = float(abs(np.dot(np.cross(m[0], m[1]), m[2])))
        return vol

    @property
    def a(self):
        """Return a lattice vector length."""
        return self.lat_lengths()[0]

    @property
    def b(self):
        """Return b lattice vector length."""
        return self.lat_lengths()[1]

    @property
    def c(self):
        """Return c lattice vector length."""
        return self.lat_lengths()[2]

    @property
    def alpha(self):
        """Return alpha lattice vector angle."""
        return self.lat_angles()[0]

    @property
    def beta(self):
        """Return beta lattice vector angle."""
        return self.lat_angles()[1]

    @property
    def gamma(self):
        """Return gamma lattice vector angle."""
        return self.lat_angles()[2]

    @property
    def abc(self):
        """Return lattice vector lengths."""
        return self.lat_lengths()

    @property
    def angles(self):
        """Return lattice vector angles."""
        return self.lat_angles()

    def lat_angles(self, tol=1e-2, radians=False):
        """Return lattice vector angles in radians or degree."""
        lengths = self.lat_lengths()
        angles = []
        for i in range(3):
            j = i - 1
            k = i - 2
            leng2 = lengths[j] * lengths[k]
            if leng2 > tol:
                tmp = np.dot(self._lat[j], self._lat[k]) / leng2
                angle = round(180.0 * np.arccos(tmp) / np.pi, 4)
            else:
                angle = 90.0
            angles.append(angle)
        if radians:
            angles = [round(angleX * np.pi / 180.0, 4) for angleX in angles]
        return angles

    @property
    def parameters(self):
        """Return lattice vector angles in radians or degree."""
        return [
            self.a,
            self.b,
            self.c,
            self.angles[0],
            self.angles[1],
            self.angles[2],
        ]

    @staticmethod
    def from_parameters(a, b, c, alpha, beta, gamma):
        """Construct Lattice from lattice parameter information."""
        angles_r = np.radians([alpha, beta, gamma])
        cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
        sin_alpha, sin_beta, sin_gamma = np.sin(angles_r)
        tmp = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
        val = abs_cap(tmp)
        gamma_star = np.arccos(val)
        vector_a = [a * sin_beta, 0.0, a * cos_beta]
        vector_b = [
            -b * sin_alpha * np.cos(gamma_star),
            b * sin_alpha * np.sin(gamma_star),
            b * cos_alpha,
        ]
        vector_c = [0.0, 0.0, float(c)]
        return Lattice(lattice_mat=[vector_a, vector_b, vector_c])

    @staticmethod
    def cubic(a):
        """Construct cubic Lattice from lattice parameter information."""
        return Lattice.from_parameters(a, a, a, 90, 90, 90)

    @staticmethod
    def tetragonal(a, c):
        """Construct tetragonal Lattice from lattice parameter information."""
        return Lattice.from_parameters(a, a, c, 90, 90, 90)

    @staticmethod
    def orthorhombic(a, b, c):
        """Construct orthorhombic Lattice."""
        return Lattice.from_parameters(a, b, c, 90, 90, 90)

    @staticmethod
    def monoclinic(a, b, c, beta):
        """Construct monoclinic Lattice from lattice parameter information."""
        return Lattice.from_parameters(a, b, c, 90, beta, 90)

    @staticmethod
    def hexagonal(a, c):
        """Construct hexagonal Lattice from lattice parameter information."""
        return Lattice.from_parameters(a, a, c, 90, 90, 120)

    @staticmethod
    def rhombohedral(a, alpha):
        """Construct rhombohedral Lattice."""
        return Lattice.from_parameters(a, a, a, alpha, alpha, alpha)

    def to_dict(self):
        """Return lattice parameter information as a dictionary."""
        d = OrderedDict()
        d["matrix"] = self.matrix
        return d

    @classmethod
    def from_dict(self, d):
        """Construct Lattice from lattice matrix dictionary."""
        return Lattice(lattice_mat=d["matrix"])

    @property
    def matrix(self):
        """Return lattice matrix."""
        return self.lattice()

    @property
    def inv_matrix(self):
        """Return inverse lattice matrix."""
        return self.inv_lattice()

    def lattice(self):
        """Return lattice matrix."""
        return self._lat

    def inv_lattice(self):
        """Return inverse lattice matrix."""
        # if self._inv_lat == None:
        self._inv_lat = np.linalg.inv(self._lat)
        return self._inv_lat

    def cart_coords(self, frac_coords):
        """Return cartesian coords from fractional coords using Lattice."""
        return np.dot(np.array(frac_coords), self._lat)

    def frac_coords(self, cart_coords):
        """Return fractional coords from cartesian coords using Lattice."""
        return np.dot(np.array(cart_coords), self.inv_lattice())

    def reciprocal_lattice(self):
        """Return reciprocal Lattice."""
        return Lattice(2 * np.pi * np.linalg.inv(self._lat).T)

    def reciprocal_lattice_crystallographic(self):
        """Return reciprocal Lattice without 2 * pi."""
        return Lattice(self.reciprocal_lattice().matrix / (2 * np.pi))

    def get_points_in_sphere(self, frac_points, center, r):
        """
        Find all points within a sphere from the point.

        Takes into account periodic boundary conditions.
        This includes sites in other periodic images.
        Adapted from pymatgen.
        """
        recp_len = np.array(self.reciprocal_lattice().lat_lengths()) / (
            2 * np.pi
        )
        nmax = float(r) * recp_len + 0.01

        # Get the fractional coordinates of the center of the sphere
        pcoords = self.frac_coords(center)
        center = np.array(center)

        # Prepare the list of output atoms
        n = len(frac_points)
        fcoords = np.array(frac_points) % 1
        indices = np.arange(n)

        # Generate all possible images that could be within `r` of `center`
        mins = np.floor(pcoords - nmax)
        maxes = np.ceil(pcoords + nmax)
        arange = np.arange(start=mins[0], stop=maxes[0], dtype=np.int)
        brange = np.arange(start=mins[1], stop=maxes[1], dtype=np.int)
        crange = np.arange(start=mins[2], stop=maxes[2], dtype=np.int)
        arange = arange[:, None] * np.array([1, 0, 0], dtype=np.int)[None, :]
        brange = brange[:, None] * np.array([0, 1, 0], dtype=np.int)[None, :]
        crange = crange[:, None] * np.array([0, 0, 1], dtype=np.int)[None, :]
        tmp_cr = crange[None, None, :]
        images = arange[:, None, None] + brange[None, :, None] + tmp_cr

        # Generate the coordinates of all atoms within these images
        tmp_img = images[None, :, :, :, :]
        shifted_coords = fcoords[:, None, None, None, :] + tmp_img

        # Determine distance from `center`
        cart_coords = self.cart_coords(fcoords)
        cart_images = self.cart_coords(images)
        tmp_img = cart_images[None, :, :, :, :]
        coords = cart_coords[:, None, None, None, :] + tmp_img
        coords -= center[None, None, None, None, :]
        coords **= 2
        d_2 = np.sum(coords, axis=4)

        # Determine which points are within `r` of `center`
        within_r = np.where(d_2 <= r ** 2)
        return (
            shifted_coords[within_r],
            np.sqrt(d_2[within_r]),
            indices[within_r[0]],
            images[within_r[1:]],
        )

    def find_all_matches(self, other_lattice, ltol=1e-5, atol=1):
        """Find all lattice mappings, adapted from pymatgen."""
        lengths = other_lattice.lat_lengths()
        angles = other_lattice.lat_angles()
        # print ('angles',angles)
        alpha = angles[0]
        beta = angles[1]
        gamma = angles[2]
        frac, dist, _, _ = self.get_points_in_sphere(
            [[0, 0, 0]], [0, 0, 0], max(lengths) * (1 + ltol)
        )
        cart = self.cart_coords(frac)
        # this can't be broadcast because they're different lengths
        inds = [
            np.logical_and(
                dist / leng < 1 + ltol, dist / leng > 1 / (1 + ltol)
            )  # type: ignore
            for leng in lengths
        ]
        c_a, c_b, c_c = (cart[i] for i in inds)
        f_a, f_b, f_c = (frac[i] for i in inds)
        l_a, l_b, l_c = (
            np.sum(c ** 2, axis=-1) ** 0.5 for c in (c_a, c_b, c_c)
        )

        def get_angles(v1, v2, l1, l2):
            x = np.inner(v1, v2) / l1[:, None] / l2
            x[x > 1] = 1
            x[x < -1] = -1
            angles = np.arccos(x) * 180.0 / np.pi
            return angles

        alphab = np.abs(get_angles(c_b, c_c, l_b, l_c) - alpha) < atol
        betab = np.abs(get_angles(c_a, c_c, l_a, l_c) - beta) < atol
        gammab = np.abs(get_angles(c_a, c_b, l_a, l_b) - gamma) < atol

        for i, all_j in enumerate(gammab):
            inds = np.logical_and(
                all_j[:, None], np.logical_and(alphab, betab[i][None, :])
            )
            for j, k in np.argwhere(inds):
                scale_m = np.array(
                    (f_a[i], f_b[j], f_c[k]), dtype=np.int
                )  # type: ignore
                if abs(np.linalg.det(scale_m)) < 1e-8:
                    continue

                aligned_m = np.array((c_a[i], c_b[j], c_c[k]))

                rotation_m = np.linalg.solve(aligned_m, other_lattice._lat)
                yield Lattice(aligned_m), rotation_m, scale_m

    def find_matches(self, other_lattice, ltol=1e-5, atol=1):
        """Find matches with length and angle tolerances."""
        for x in self.find_all_matches(other_lattice, ltol, atol):
            return x
        return None

    def _calculate_lll(self, delta=0.75):
        """
        Perform a Lenstra-Lenstra-Lovasz lattice basis reduction.

        Obtain a c-reduced basis.
        This method returns a basis which is as "good" as
        possible, with "good" defined by orthongonality of the lattice vectors.
        This basis is used for all the periodic boundary condition calcs.
        Adapted from pymatgen.

        Args:

            delta (float): Reduction parameter.
                           Default of 0.75 is usually fine.

        Returns:
            Reduced lattice matrix, mapping to get to that lattice.
        """
        # Transpose the lattice matrix first so that basis vectors are columns.
        # Makes life easier.
        a = self._lat.copy().T

        b = np.zeros((3, 3))  # Vectors after the Gram-Schmidt process
        u = np.zeros((3, 3))  # Gram-Schmidt coeffieicnts
        m = np.zeros(3)  # These are the norm squared of each vec.

        b[:, 0] = a[:, 0]
        m[0] = dot(b[:, 0], b[:, 0])
        for i in range(1, 3):
            u[i, 0:i] = dot(a[:, i].T, b[:, 0:i]) / m[0:i]
            b[:, i] = a[:, i] - dot(b[:, 0:i], u[i, 0:i].T)
            m[i] = dot(b[:, i], b[:, i])

        k = 2

        mapping = np.identity(3, dtype=np.double)
        while k <= 3:
            # Size reduction.
            for i in range(k - 1, 0, -1):
                q = round(u[k - 1, i - 1])
                if q != 0:
                    # Reduce the k-th basis vector.
                    a[:, k - 1] = a[:, k - 1] - q * a[:, i - 1]
                    mapping[:, k - 1] = (
                        mapping[:, k - 1] - q * mapping[:, i - 1]
                    )
                    uu = list(u[i - 1, 0 : (i - 1)])
                    uu.append(1)
                    # Update the GS coefficients.
                    u[k - 1, 0:i] = u[k - 1, 0:i] - q * np.array(uu)

            # Check the Lovasz condition.
            if dot(b[:, k - 1], b[:, k - 1]) >= (
                delta - abs(u[k - 1, k - 2]) ** 2
            ) * dot(b[:, (k - 2)], b[:, (k - 2)]):
                # Increment k if the Lovasz condition holds.
                k += 1
            else:
                # If the Lovasz condition fails,
                # swap the k-th and (k-1)-th basis vector
                v = a[:, k - 1].copy()
                a[:, k - 1] = a[:, k - 2].copy()
                a[:, k - 2] = v

                v_m = mapping[:, k - 1].copy()
                mapping[:, k - 1] = mapping[:, k - 2].copy()
                mapping[:, k - 2] = v_m

                # Update the Gram-Schmidt coefficients
                for s in range(k - 1, k + 1):
                    u[s - 1, 0 : (s - 1)] = (
                        dot(a[:, s - 1].T, b[:, 0 : (s - 1)]) / m[0 : (s - 1)]
                    )
                    b[:, s - 1] = a[:, s - 1] - dot(
                        b[:, 0 : (s - 1)], u[s - 1, 0 : (s - 1)].T
                    )
                    m[s - 1] = dot(b[:, s - 1], b[:, s - 1])

                if k > 2:
                    k -= 1
                else:
                    # We have to do p/q, so do lstsq(q.T, p.T).T instead.
                    p = dot(a[:, k:3].T, b[:, (k - 2) : k])
                    q = np.diag(m[(k - 2) : k])
                    result = np.linalg.lstsq(q.T, p.T, rcond=None)[
                        0
                    ].T  # type: ignore
                    u[k:3, (k - 2) : k] = result

        return a.T, mapping.T

    def get_lll_reduced_lattice(self, delta=0.75):
        """
        Get LLL reduced lattice.

        Adpted from pymatgen.
        Args:

           delta: Delta parameter.

        Returns:
             LLL reduced Lattice.
        """
        if delta not in self._lll_matrix_mappings:
            self._lll_matrix_mappings[delta] = self._calculate_lll()
        return Lattice(self._lll_matrix_mappings[delta][0])


def lattice_coords_transformer(
    old_lattice_mat=[], new_lattice_mat=[], cart_coords=[]
):
    """Transform coords to a new lattice."""
    M = np.linalg.solve(new_lattice_mat, old_lattice_mat)
    #  Maintains the z-distances - kfg
    M[2, 2] = 1.0
    new_cart_coords = np.dot(cart_coords, M)
    return new_cart_coords


"""
if __name__ == "__main__":

    box = [[10, 0, 0], [0, 10, 0], [0, 0, 10]]
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    lat = Lattice(box)
    print("lll", lat._calculate_lll())
    print("lll_educed", lat.get_lll_reduced_lattice()._lat)
    frac_coords = [[0, 0, 0], [0.5, 0.5, 0.5]]
    print(lat.cart_coords(frac_coords)[1][1])

    cart_coords = [[0, 0, 0], [5, 5, 5]]
    print(lat.frac_coords(cart_coords)[1][1])
"""



import os

import tarfile

example_fold_tgz = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "examples",
    "lammps",
    "Al03.eam.alloy_nist.tgz",
)


example_fold = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "examples",
    "lammps",
    "Al03.eam.alloy_nist",
)

if not os.path.isdir(example_fold):
    tar = tarfile.open(example_fold_tgz)
    tar.extractall(example_fold)
    tar.close()



lammps_path = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "..",
    "examples",
    "lammps",
    "Al03.eam.alloy_nist",
    "Al03.eam.alloy_nist",
    "bulk@mp-134_fold",
)

cwd = str(os.getcwd())


def test_lammps():
    os.chdir(lammps_path)
    cmd = str("python master.py")
    os.system(cmd)


os.chdir(cwd)



"""
Coulomb matrix for Atoms.

Refer to: 10.1103/PhysRevLett.108.058301
"""

import numpy as np
from jarvis.core.specie import Specie


def coulomb_matrix(atoms="", max_dim=100):
    """Convert Atoms class to max_dim x max_dim matrix.

    Args:

        atoms: atoms object

        max_dim: maximum number of atoms=sqrt(max_dim)

    Returns:
          z: numpy array of 1 x max_dim dimension
    """
    natoms = atoms.num_atoms
    mat = np.zeros((natoms, natoms))

    for ii, i in enumerate(atoms.elements):
        for jj, j in enumerate(atoms.elements):
            if ii == jj:
                mat[ii, jj] = 0.5 * Specie(i).Z ** 2.4
            else:
                a = atoms.cart_coords[ii]
                b = atoms.cart_coords[jj]
                dist = np.linalg.norm(a - b)
                mat[ii, jj] = (Specie(i).Z * Specie(j).Z) / dist
    tmp = mat.ravel()
    if max_dim < len(tmp):
        print("WARNING: Increase max_dim")
    padding = max_dim - len(tmp)
    z = np.pad(tmp, (0, padding), "constant")
    return z


"""
if __name__ == "__main__":
    from jarvis.core.atoms import Atoms
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    z = coulomb_matrix(Si)
    print(z, len(z))
"""



"""Methods for handling input/output files for pennylane."""
# Reference: https://arxiv.org/abs/2102.11452
from pennylane.utils import decompose_hamiltonian
import pennylane as qml
import numpy as np
from jarvis.io.qiskit.inputs import HermitianSolver


def run_vqe(mat=[], max_iterations=100, conv_tol=1e-04, step_size=0.01):
    """Run pennylane VQE solver."""
    herm = HermitianSolver(mat)
    hk = herm.mat
    nwires = herm.n_qubits()
    print("nwires=", nwires)
    coeff, obs_list = decompose_hamiltonian(hk)
    H = qml.Hamiltonian(coeff, obs_list)
    # print(H)
    ansatz = qml.templates.StronglyEntanglingLayers
    dev = qml.device("default.qubit", wires=nwires)
    qml.enable_tape()
    cost_fn = qml.ExpvalCost(ansatz, H, dev)
    opt = qml.GradientDescentOptimizer(stepsize=step_size)
    params = np.random.rand(
        1, nwires, 3
    )  # torch.rand([2, nwires, 3]).cpu().detach().numpy() #init_params
    prev_energy = cost_fn(params)
    gd_param_history = [params]
    gd_cost_history = [prev_energy]
    for n in range(max_iterations):
        # Take step
        params = opt.step(cost_fn, params)
        gd_param_history.append(params)
        # print (n,ansatz)
        # print ()
        # Compute energy
        energy = cost_fn(params)
        gd_cost_history.append(energy)
        # Calculate difference between new and old energies
        conv = np.abs(energy - prev_energy)
        if n % 20 == 0:
            print(
                "Iteration={:}, Energy={:.8f} Ha,Convergence parameter = {"
                ":.8f} Ha".format(n, energy, conv)
            )
        if conv <= conv_tol:
            break
        prev_energy = energy

    # print()
    print("Final value of the energy = {:.8f} Ha".format(energy))
    # print("Number of iterations = ", n)
    vect = np.array(cost_fn.hamiltonian.coeffs).reshape(hk.shape)
    return energy, vect


"""
if __name__ == "__main__":
    from jarvis.db.figshare import (
        get_wann_phonon,
        get_hk_tb,
        get_wann_electron,
    )

    w, atoms, o = get_wann_electron("JVASP-816")
    # w, atoms = get_wann_phonon("JVASP-816")
    hk = get_hk_tb(w=w, k=[0.0, 0.0, 0])
    run_vqe(mat=hk)
"""



import glob
import os

from setuptools import setup, find_packages

JARVIS_DIR = os.path.dirname(os.path.abspath(__file__))

base_dir = os.path.dirname(__file__)
with open(os.path.join(base_dir, "README.rst")) as f:
    long_d = f.read()

setup(
    name="jarvis-tools",
    version="2021.3.13",
    long_description=long_d,
    install_requires=[
        "numpy>=1.18.5",
        "scipy>=1.4.1",
        "matplotlib>=3.0.0",
        "spglib>=1.14.1",
        "joblib>=0.14.1",
        "requests>=2.23.0",
        "toolz>=0.9.0",
        "xmltodict>=0.11.0",
        "tqdm>=4.59.0",
    ],
    package_data={
        "jarvis.core": ["Elements.json", "element_charge.json"],
        "jarvis.tasks.lammps.templates": [
            "displace.mod",
            "inelastcomb.mod",
            "inelast_min.mod",
            "inelast.mod",
            "inelast_nobox.mod",
            "inelastreax.mod",
            "relax.mod",
            "run0.mod",
        ],
        "jarvis.io.vasp": ["default_potcars.json"],
        "jarvis.analysis.solarefficiency": ["am1.5G.dat"],
        "jarvis.io.wannier": ["default_semicore.json"],
        "jarvis.analysis.diffraction": ["atomic_scattering_params.json"],
        "jarvis": ["LICENSE.rst"],
    },
    extras_require={
        "ai": [
            "torch",
            "keras",
            "tensorflow",
            "scikit-learn",
            "flask",
            "pandas",
        ],
        "babel": ["openbabel", "pybel"],
        "doc": ["sphinx>=1.3.1", "sphinx-rtd-theme>=0.1.8"],
    },
    author="Kamal Choudhary",
    author_email="kamal.choudhary@nist.gov",
    description=(
        "jarvis-tools: an open-source software package for data-driven atomistic materials design. https://jarvis.nist.gov/"
    ),
    license="NIST",
    url="https://github.com/usnistgov/jarvis",
    packages=find_packages(),
    # long_description=open(os.path.join(os.path.dirname(__file__), "README.rst")).read(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    # scripts=glob.glob(os.path.join(JARVIS_DIR,  "*"))
)



"""Modules related to chemistry of periodic-table elements."""

import os
import json
import numpy as np
import functools

el_chem_json_file = str(
    os.path.join(os.path.dirname(__file__), "Elements.json")
)
el_chem_json = open(el_chem_json_file, "r")
chem_data = json.load(el_chem_json)
el_chem_json.close()

el_chrg_json_file = str(
    os.path.join(os.path.dirname(__file__), "element_charge.json")
)
el_chrg_json = open(el_chrg_json_file, "r")
chrg_data = json.load(el_chrg_json)
el_chrg_json.close()
cgcnn_feature_json = os.path.join(os.path.dirname(__file__), "atom_init.json")


def get_descrp_arr_name(elm="Al"):
    """
    Get chemical descriptors for an element.

    Can be used in JARVIS-ML.

    Args:

           elm: element name

    Returns:
             arr: array value
    """
    arr = []
    try:
        dat = chem_data
        d = dat[elm]
        arr = []
        for k, v in d.items():
            arr.append(k)
    except Exception:
        pass
    return arr


class Specie(object):
    """
    Specie object for chemistry information.

    Used in defining chemistry of a material.

    >>> el = Specie('Al')
    >>> el.Z
    13
    >>> round(el.atomic_mass,2)
    26.98
    >>> el.symbol
    'Al'
    >>> round(el.get_chgdescrp_arr[1],2)
    12.17
    >>> round(el.get_descrp_arr[1],2)
    2792.11
    >>> el = Specie('asdfg')
    >>> el.element_property("asdfg")
    nan

    """

    def __init__(self, symbol=""):
        """Initialize with periodic table element."""
        self.symbol = symbol
        self._data = chem_data

    @property
    def Z(self):
        """Get atomic number."""
        return self.element_property("Z")

    @property
    def atomic_mass(self):
        """Get atomic mass."""
        return self.element_property("atom_mass")

    @property
    def get_chgdescrp_arr(self):
        """
        Get charge descriptors for an element.

        Gives 378 array data.

        Args:

           elm: element name

        Returns:
             arr: array value
        """
        arr = []

        # try:
        arr = chrg_data[self.symbol][0][1]
        # except:
        #    pass
        return arr

    @property
    def get_descrp_arr(self):
        """
        Get chemical descriptors for an element.

        Gives 438 array data.

        Args:

           elm: element name

        Returns:
             arr: array value
        """
        arr = []

        d = chem_data[self.symbol]
        arr = []
        for k, v in d.items():
            arr.append(v)
        arr = np.array(arr).astype(float)
        return arr

    @property
    def atomic_rad(self):
        """Get atomic radii."""
        return self.element_property("atom_rad")

    def element_property(self, key=""):
        """
        Get element property from the list of keys.

        These 84 keys are:
        keys = [
                "is_halogen",
                "row",
                "GV",
                "nfunfill",
                "C-9",
                "C-8",
                "C-7",
                "C-6",
                "C-5",
                "C-4",
                "C-3",
                "C-2",
                "C-1",
                "C-0",
                "me1",
                "me3",
                "me2",
                "max_oxid_s",
                "npvalence",
                "mp",
                "first_ion_en",
                "ndunfill",
                "op_eg",
                "jv_enp",
                "nfvalence",
                "polzbl",
                "oq_bg",
                "atom_rad",
                "atom_mass",
                "is_alkali",
                "C-13",
                "C-12",
                "C-11",
                "C-10",
                "C-17",
                "C-16",
                "C-15",
                "C-14",
                "C-19",
                "C-18",
                "voro_coord",
                "is_noble_gas",
                "e1",
                "e3",
                "e2",
                "is_lanthanoid",
                "ndvalence",
                "KV",
                "min_oxid_s",
                "nsunfill",
                "C-26",
                "X",
                "is_actinoid",
                "C-28",
                "C-29",
                "C-27",
                "C-24",
                "C-25",
                "C-22",
                "C-23",
                "C-20",
                "C-21",
                "avg_ion_rad",
                "nsvalence",
                "is_metalloid",
                "elec_aff",
                "coulmn",
                "mol_vol",
                "bp",
                "C-31",
                "C-30",
                "C-33",
                "C-32",
                "C-35",
                "C-34",
                "is_transition_metal",
                "block",
                "therm_cond",
                "Z",
                "is_alkaline",
                "npunfill",
                "oq_enp",
                "mop_eg",
                "hfus",
            ]
        """
        val = np.nan
        try:
            val = self._data[self.symbol][key]
        except Exception:
            pass
        return val


BASIC_FEATURES = [
    "Z",
    "coulmn",
    "row",
    "X",
    "atom_rad",
    "nsvalence",
    "npvalence",
    "ndvalence",
    "nfvalence",
    "first_ion_en",
    "elec_aff",
]


@functools.lru_cache(maxsize=None)
def get_node_attributes(species, atom_features="atomic_number"):
    """Get specific node features for an element."""
    feature_sets = ("atomic_number", "basic", "cfid", "cgcnn")

    if atom_features not in feature_sets:
        raise NotImplementedError(
            f"atom features must be one of {feature_sets}"
        )

    if atom_features == "cfid":
        return Specie(species).get_descrp_arr
    elif atom_features == "atomic_number":
        return [Specie(species).element_property("Z")]
    elif atom_features == "basic":
        return [
            Specie(species).element_property(prop) for prop in BASIC_FEATURES
        ]
    elif atom_features == "cgcnn":
        # load from json, key by atomic number
        key = str(Specie(species).element_property("Z"))
        with open(cgcnn_feature_json, "r") as f:
            i = json.load(f)
        return i[key]


"""
if __name__ == "__main__":
    el = Specie("Al")
    #print(el.get_chgdescrp_arr)
    print(len(el.get_descrp_arr))
    #print (get_descrp_arr_name())
"""



"""
Class for reading wannier outouts.

Such as wannier90.wout and wannier90_hr.dat
"""

import os
import json
import matplotlib.pyplot as plt
import cmath
import numpy as np
import math
import time
from jarvis.core.kpoints import Kpoints3D
from jarvis.io.vasp.outputs import Vasprun
import scipy
import scipy.optimize as opt
import copy as copy


def get_projectors_for_formula(
    semi_core_states=None, formula_dict={"Bi": 2, "Se": 3}
):
    """Get semi core states from formula dict."""
    if semi_core_states is None:
        path_semi_core = str(
            os.path.join(os.path.dirname(__file__), "default_semicore.json")
        )
        f = open(path_semi_core, "r")
        semi_core_states = json.load(f)
        f.close()
    arr = []
    for i, j in formula_dict.items():
        dat = semi_core_states[i]
        arr.append([i, j, [str(k) for k in dat[2].split(",")]])
    return arr


def get_orbitals(
    projection_info=[["Bi", 2, ["s", "p"]], ["Se", 3, ["s", "p"]]],
    desired_orbitals=[["Bi", "p"]],
    soc=True,
    ncells=1,
    supercell=[1, 1, 6],
    surfaceonly=False,
):
    """Get spdf orbitals."""
    # projection_info example for Bi2Se3 with s and p orbital projections
    # [["Bi", 2, ["s","p"]], ["Se", 3, ["s","p"]]]

    # orbitals wanted example
    # [["Bi"]]   all Bi orbitals

    # [["Bi", "p"]] all Bi p orbitals

    # [["Bi", "px"], ["Bi" ,"py"]] all Bi px, py oribitals

    # [["Bi", "s"], ["Se", "s"]]  Bi s and Se s

    # so for spin-orbit

    c = 0

    projection_dict = {}

    # print "get_orbs ", so
    # print projection_info

    for proj in projection_info:

        atom = proj[0]
        natom = proj[1]
        orbitals = proj[2]

        for n in range(natom):
            for o in orbitals:
                if o == "s":
                    if (atom, "s") not in projection_dict:
                        projection_dict[(atom, "s")] = []
                    projection_dict[(atom, "s")].append(c)
                    c += 1

                elif o == "p":
                    if (atom, "p") not in projection_dict:
                        projection_dict[(atom, "p")] = []
                        projection_dict[(atom, "pz")] = []
                        projection_dict[(atom, "py")] = []
                        projection_dict[(atom, "px")] = []

                    projection_dict[(atom, "p")].append(c)
                    projection_dict[(atom, "pz")].append(c)
                    c += 1

                    projection_dict[(atom, "p")].append(c)
                    projection_dict[(atom, "px")].append(c)
                    c += 1

                    projection_dict[(atom, "p")].append(c)
                    projection_dict[(atom, "py")].append(c)
                    c += 1

                elif o == "d":
                    if (atom, "p") not in projection_dict:
                        projection_dict[(atom, "d")] = []
                        projection_dict[(atom, "dz2")] = []
                        projection_dict[(atom, "dxz")] = []
                        projection_dict[(atom, "dyz")] = []
                        projection_dict[(atom, "dx2y2")] = []
                        projection_dict[(atom, "dxy")] = []

                        projection_dict[(atom, "d")].append(c)
                        projection_dict[(atom, "dz2")].append(c)
                        c += 1

                        projection_dict[(atom, "d")].append(c)
                        projection_dict[(atom, "dxz")].append(c)
                        c += 1

                        projection_dict[(atom, "d")].append(c)
                        projection_dict[(atom, "dyz")].append(c)
                        c += 1

                        projection_dict[(atom, "d")].append(c)
                        projection_dict[(atom, "dx2y2")].append(c)
                        c += 1

                        projection_dict[(atom, "d")].append(c)
                        projection_dict[(atom, "dxy")].append(c)
                        c += 1

    nwan = c
    if soc:
        for (atom, orb) in projection_dict.keys():
            new_ind = []
            for i in projection_dict[(atom, orb)]:
                new_ind.append(i + nwan)
            projection_dict[(atom, orb)] += new_ind
        nwan = nwan * 2
    print("nwan = ", nwan)

    inds = []
    for d in desired_orbitals:
        if len(d) == 1:
            for orb in ["s", "p", "d"]:
                if (d[0], orb) in projection_dict:
                    new_orbs = projection_dict[(d[0], orb)]
                    inds += new_orbs
        else:
            new_orbs = projection_dict[tuple(d)]
            inds += new_orbs

        if ncells > 1 and not surfaceonly:

            inds_super = []
            for n in range(ncells):
                for i in inds:
                    inds_super.append(i + n * nwan)

            return inds_super

        elif ncells > 1 and surfaceonly:
            inds_super = []
            for n in [0, ncells - 1]:
                for i in inds:
                    inds_super.append(i + n * nwan)

            return inds_super

    return inds


class WannierHam(object):
    """Construct WannierHamltonian object."""

    def __init__(
        self,
        filename="wannier90_hr.dat",
        nwan=None,
        nr=None,
        sym_r=None,
        H_int=None,
        H_val=None,
        H=None,
        HR=None,
    ):
        """Initialize the class."""
        self.filename = filename
        self.nr = nr
        self.nwan = nwan
        self.sym_r = sym_r
        self.H_int = H_int
        self.H_val = H_val
        self.H = H
        self.HR = HR

        if self.nr is None:
            self.read_ham()

    def to_dict(self):
        """Convert to dictionary."""
        info = {}
        info["nr"] = self.nr
        info["filename"] = self.filename
        info["nwan"] = self.nwan
        info["sym_r"] = self.sym_r
        info["H_int"] = self.H_int
        info["H_val"] = self.H_val
        info["H"] = self.H
        info["HR"] = self.HR
        return info

    def find_nodes(
        self,
        num_occ=8,
        origin=[-0.5, -0.5, -0.5],
        k1=[1, 0, 0],
        k2=[0, 1, 0],
        k3=[0, 0, 1],
        nk1=10,
        nk2=10,
        nk3=10,
        sig=[0.01, 0.02, 0.05, 0.1, 0.2, 0.3],
        thresh=-10,
        node_tol=0.001,
        use_min=True,
    ):
        """Find nodes, Dirac, Wyel points."""
        K = np.zeros((nk1 * 4, nk2 * 4, nk3 * 4, 3), dtype=float)
        k1 = np.array(k1, dtype=float)
        k2 = np.array(k2, dtype=float)
        k3 = np.array(k3, dtype=float)

        for c1 in range(4 * nk1):
            for c2 in range(4 * nk2):
                for c3 in range(4 * nk3):
                    K[c1, c2, c3, :] = (
                        origin
                        + k1 * float(c1) / float(nk1 * 4)
                        + k2 * float(c2) / float(nk2 * 4)
                        + k3 * float(c3) / float(nk3 * 4)
                    )

                    #        VALS = np.zeros((nk1,nk2, ham.nwan))

        IMAGE = np.zeros((nk1 * 4, nk2 * 4, nk3 * 4, len(sig)))
        DIRECTGAP = np.zeros((nk1 * 4, nk2 * 4, nk3 * 4))
        DIRECTGAP[:, :, :] = 100.0

        VAL = np.ones((nk1 * 4, nk2 * 4, nk3 * 4)) * 100.0

        #        sig_0 = 0.2

        # initial pass

        for c1 in range(0, 4 * nk1, 4):
            print("c1 ", c1)
            for c2 in range(0, 4 * nk2, 4):
                for c3 in range(0, 4 * nk3, 4):
                    k = K[c1, c2, c3, :]
                    val, vect, p = self.solve_ham(k, proj=None)

                    d = val[num_occ] - val[num_occ - 1]
                    DIRECTGAP[c1, c2, c3] = d

                    # IMAGE[c1,c2,c3,-1] = np.exp( -(d)**2 / sig[-1]**2)

        if thresh == -10:
            thresh = max(np.min(DIRECTGAP) * 1.5, 0.03)
            print("set thresh ", thresh)

            # second pass

            #        maxval = np.max(IMAGE[:,:,:,-1])
            #        thresh = maxval * 0.1
            #        print("thresh ", maxval, thresh)

        def fun_to_min(k):
            val, vect, p = self.solve_ham(k, proj=None)
            return val[num_occ] - val[num_occ - 1]

        num_thresh = 0

        min_gap = 1000000000.0
        min_cond = 1000000000.0
        max_val = -1000000000.0

        dk1 = float(1) / float(nk1 * 4) / 2.0
        dk2 = float(1) / float(nk2 * 4) / 2.0
        dk3 = float(1) / float(nk3 * 4) / 2.0

        weyl_points = []
        dirac_points = []
        higher_order_points = []

        for c1 in range(0, 4 * nk1, 4):
            print("c1 ", c1)
            for c2 in range(0, 4 * nk2, 4):
                for c3 in range(0, 4 * nk3, 4):
                    for c1a in range(4):
                        for c2a in range(4):
                            for c3a in range(4):

                                if c1a > 0:
                                    c1p = (c1 + 4) % (4 * nk1)
                                else:
                                    c1p = c1
                                if c2a > 0:
                                    c2p = (c2 + 4) % (4 * nk2)
                                else:
                                    c2p = c2
                                if c3a > 0:
                                    c3p = (c3 + 4) % (4 * nk3)
                                else:
                                    c3p = c3

                                if (
                                    DIRECTGAP[c1, c2, c3] < thresh
                                    or DIRECTGAP[c1, c2, c3p] < thresh
                                    or DIRECTGAP[c1, c2p, c3] < thresh
                                    or DIRECTGAP[c1, c2p, c3p] < thresh
                                    or DIRECTGAP[c1p, c2, c3] < thresh
                                    or DIRECTGAP[c1p, c2, c3p] < thresh
                                    or DIRECTGAP[c1p, c2p, c3] < thresh
                                    or DIRECTGAP[c1p, c2p, c3p] < thresh
                                ):
                                    num_thresh += 1
                                    k = K[c1 + c1a, c2 + c2a, c3 + c3a, :]

                                    if use_min:
                                        B = opt.Bounds(
                                            k - [dk1, dk2, dk3],
                                            k + [dk1, dk2, dk3],
                                            keep_feasible=True,
                                        )
                                        val, vect, p = self.solve_ham(
                                            k, proj=None
                                        )
                                        if (
                                            val[num_occ] - val[num_occ - 1]
                                            < 0.15
                                        ):
                                            res = opt.minimize(
                                                fun_to_min,
                                                k,
                                                method="Powell",
                                                bounds=B,
                                                tol=0.5e-3,
                                                options={"maxfev": 5},
                                            )
                                            if res.fun < 0.005:
                                                res = opt.minimize(
                                                    fun_to_min,
                                                    res.x,
                                                    method="Powell",
                                                    bounds=B,
                                                    tol=1e-5,
                                                    options={"maxfev": 30},
                                                )
                                            val, vect, p = self.solve_ham(
                                                res.x, proj=None
                                            )

                                            k = res.x
                                    else:
                                        val, vect, p = self.solve_ham(
                                            k, proj=None
                                        )

                                    d = val[num_occ] - val[num_occ - 1]
                                    if d < node_tol:
                                        # degen = 0
                                        gap_mean = (
                                            val[num_occ] + val[num_occ - 1]
                                        ) / 2.0
                                        count = np.sum(
                                            np.abs(val - gap_mean)
                                            < node_tol * 2
                                        )
                                        if count == 2:
                                            weyl_points.append(copy.copy(k))
                                            print("weyl point ", k, d, count)
                                        elif count == 4 or count == 4:
                                            dirac_points.append(copy.copy(k))
                                            print("dirac point ", k, d, count)
                                        else:
                                            higher_order_points.append(
                                                copy.copy(k)
                                            )
                                            print(
                                                "nontrivial point ",
                                                k,
                                                d,
                                                count,
                                            )

                                    DIRECTGAP[c1 + c1a, c2 + c2a, c3 + c3a] = d
                                    VAL[c1 + c1a, c2 + c2a, c3 + c3a] = 0.5 * (
                                        val[num_occ] + val[num_occ - 1]
                                    )
                                    min_gap = min(d, min_gap)
                                    min_cond = min(min_cond, val[num_occ])
                                    max_val = max(max_val, val[num_occ - 1])

        for (cs, s) in enumerate(sig):
            IMAGE[:, :, :, cs] = np.exp(-((DIRECTGAP) ** 2) / s ** 2)

        print(
            "num thresh ", num_thresh, " out of ", nk1 * nk2 * nk3 * 4 * 4 * 4
        )
        print(
            "min direct gap ", min_gap, " indirect gap  ", min_cond - max_val
        )

        print("weyl ", weyl_points)
        print("dirac ", dirac_points)
        print("higher order ", higher_order_points)

        return (
            IMAGE,
            DIRECTGAP,
            VAL,
            [min_gap, min_cond - max_val],
            weyl_points,
            dirac_points,
            higher_order_points,
        )

    def chern_number_simple(
        self,
        nocc=8,
        k1=[0, 0, 0],
        k2=[0, 0, 0],
        nk1=20,
        nk2=20,
        Kmat=None,
        usemod=True,
    ):
        """Calculate Chern number."""
        if Kmat is None:

            K = np.zeros((nk1, nk2, 3), dtype=float)

            k1 = np.array(k1, dtype=float)
            k2 = np.array(k2, dtype=float)

            #        print( 'K')

            for c1 in range(nk1):
                for c2 in range(nk2):
                    K[c1, c2, :] = k1 * float(c1) / float(nk1) + k2 * float(
                        c2
                    ) / float(nk2)

        else:
            K = Kmat

        if usemod:

            lim1 = nk1
            lim2 = nk2

        else:

            lim1 = nk1 + 1
            lim2 = nk2 + 1

        VECT = np.zeros((lim1, lim2, self.nwan, self.nwan), dtype=complex)
        gap_min = 100000000000.0
        val_max = -10000000000000.0
        cond_min = 10000000000.0

        # cond_min_p1 = 10000000000.0

        # gap_min2 = 100000000000.0

        # nwan = self.nwan

        for c1 in range(lim1):
            for c2 in range(lim2):
                k = K[c1, c2, :]
                val, vect, p = self.solve_ham(k, proj=None)
                VECT[c1, c2, :, :] = vect
                if (val[nocc] - val[nocc - 1]) < gap_min:
                    gap_min = val[nocc] - val[nocc - 1]

                if val[nocc - 1] > val_max:
                    val_max = val[nocc - 1]

                if val[nocc] < cond_min:
                    cond_min = val[nocc]

        print(
            "minimum_direct_gap", gap_min, "indirect_gap", cond_min - val_max
        )
        direct_gap = gap_min
        indirect_gap = cond_min - val_max

        Chern = 0.0
        for c1 in range(nk1):
            for c2 in range(nk2):

                if usemod:
                    c1p = (c1 + 1) % nk1
                    c2p = (c2 + 1) % nk2
                else:
                    c1p = c1 + 1
                    c2p = c2 + 1

                klist = [[c1, c2], [c1p, c2], [c1p, c2p], [c1, c2p], [c1, c2]]

                phi = 0.0
                for i in range(4):
                    M = np.dot(
                        VECT[klist[i][0], klist[i][1], :, 0:nocc].T.conj(),
                        VECT[klist[i + 1][0], klist[i + 1][1], :, 0:nocc],
                    )
                    phi = phi - (cmath.log(np.linalg.det(M))).imag

                if phi > np.pi:
                    phi = phi - 2 * np.pi
                elif phi < -np.pi:
                    phi = phi + 2 * np.pi

                if abs(phi) > 0.75:
                    print(
                        "WARNING in chern calculation, phi",
                        c1,
                        c2,
                        phi,
                        " try more k-points",
                    )

                Chern += phi / (2.0 * np.pi)

        if usemod:
            return Chern, direct_gap, indirect_gap
        else:
            return Chern, direct_gap, val_max, cond_min

    @classmethod
    def from_dict(self, info):
        """Load from dictionary."""
        w = WannierHam(
            nr=info["nr"],
            nwan=info["nwan"],
            sym_r=info["sym_r"],
            H_int=info["H_int"],
            H_val=info["H_val"],
            H=info["H"],
            HR=info["HR"],
        )
        return w

    def read_ham(self):
        """Read _hr.dat file.."""
        f = open(self.filename, "r")
        lines = f.read().splitlines()
        # allines = f.readlines()
        f.close()
        self.nwan = int(lines[1])
        self.nr = int(lines[2])
        if self.nr % 15 == 0:
            lines_r = int(math.floor(self.nr // 15))
        else:
            lines_r = int(math.floor(self.nr // 15) + 1)

        # print ('self.nwan,self.nr',self.nwan,self.nr,lines_r)
        self.sym_r = np.zeros(self.nr, dtype=float)
        # load sym ops
        for i in range(lines_r):

            num = 3 + i
            start = i * 15
            end = (i + 1) * 15
            if end > self.nr:
                end = self.nr
            self.sym_r[start:end] = [float(j) for j in lines[num].split()]
        # print (self.sym_r)

        tot = self.nwan ** 2 * self.nr
        self.H_int = np.zeros((tot, 5), dtype=int)
        self.H_val = np.zeros(tot, dtype=complex)

        c = 0
        rnum = 0
        for i in range(lines_r + 3, lines_r + 3 + tot):

            rnum = (c) // self.nwan ** 2
            # print ('lines[i].split()[0:5]',c,self.nwan**2)#,self.sym_r[rnum])
            self.H_int[c, :] = [int(j) for j in lines[i].split()[0:5]]
            # print ('lines[i][5]',lines[i])
            self.H_val[c] = (
                float(lines[i].split()[5]) + 1j * float(lines[i].split()[6])
            ) / float(self.sym_r[rnum])
            c += 1

        # print (self.H_int[0,:])
        # print (self.H_val[0])

        # print (self.H_int[-1,:])
        # print (self.H_val[-1])

        # print ('loaded ', self.filename)
        # print ('nwan: ', self.nwan)
        # print ('nr:   ', self.nr)
        # print ()

        # reshape

        nx1 = np.min(self.H_int[:, 0])
        nx2 = np.max(self.H_int[:, 0])
        ny1 = np.min(self.H_int[:, 1])
        ny2 = np.max(self.H_int[:, 1])
        nz1 = np.min(self.H_int[:, 2])
        nz2 = np.max(self.H_int[:, 2])

        self.ind = [[nx1, nx2], [ny1, ny2], [nz1, nz2]]

        ix = nx2 - nx1 + 1
        iy = ny2 - ny1 + 1
        iz = nz2 - nz1 + 1

        print("H size", ix, iy, iz, self.nwan, self.nwan)

        self.H = np.zeros((ix, iy, iz, self.nwan, self.nwan), dtype=complex)

        self.ind_dict = {}
        for i in range(self.H_val.shape[0]):
            ind = self.get_ind(self.H_int[i, 0:3])

            self.ind_dict[(ind[0], ind[1], ind[2])] = i

            nw1 = self.H_int[i, 3]
            nw2 = self.H_int[i, 4]

            self.H[ind[0], ind[1], ind[2], nw1 - 1, nw2 - 1] = self.H_val[i]

        # print ('done reshaping1')
        nr = ix * iy * iz
        self.R = np.zeros((nr, 3), dtype=float)
        self.HR = np.zeros((nr, self.nwan ** 2), dtype=complex)

        c = 0
        for x in range(nx1, nx2 + 1):
            for y in range(ny1, ny2 + 1):
                for z in range(nz1, nz2 + 1):
                    #                    ind = self.ind_dict[(x,y,z)]
                    ind = self.get_ind([x, y, z])
                    self.R[c, :] = [x, y, z]
                    self.HR[c, :] = np.reshape(
                        self.H[ind[0], ind[1], ind[2], :, :], self.nwan ** 2
                    )
                    c += 1

        if c != nr:
            ValueError("c is not equal to r", c, nr)

        return self.R, self.H, self.HR

    def get_ind(self, nxyz):
        """Get index."""
        return [
            nxyz[0] - self.ind[0][0],
            nxyz[1] - self.ind[1][0],
            nxyz[2] - self.ind[2][0],
        ]

    def solve_ham(self, k=[0, 0, 0], proj=None):
        """Solve Wannier Hamiltonian at a k-point."""
        nr = self.R.shape[0]
        # print ('nr==',nr,self.nr)
        hk = np.zeros((self.nwan, self.nwan), dtype=complex)

        kmat = np.tile(k, (nr, 1))

        exp_ikr = np.exp(1.0j * 2 * np.pi * np.sum(kmat * self.R, 1))

        temp = np.zeros(self.nwan ** 2, dtype=complex)
        for i in range(nr):
            temp += exp_ikr[i] * self.HR[i, :]

        hk = np.reshape(temp, (self.nwan, self.nwan))

        hk = (hk + hk.T.conj()) / 2.0

        val, vect = np.linalg.eigh(hk)

        if proj is not None:
            p = np.real(np.sum(vect[proj, :] * np.conj(vect[proj, :]), 0))
        else:
            p = np.ones(val.shape)

        #        print 'proj', np.sum(np.sum(p))

        return val.real, vect, p

    def band_structure_eigs(self, kpath=None, proj=None, efermi=0.0):
        """Get eigenvalues for band eigenvalues."""
        eigs = []
        for i in kpath:
            # print (i)
            val, vect, p = self.solve_ham(k=i, proj=proj)
            eigs.append(val - efermi)
        return np.array(eigs)

    def get_bandstructure_plot(
        self, atoms=None, efermi=0.0, filename="bs.png", yrange=[-4, 4]
    ):
        """Get bandstructure plot.."""
        kpoints, labels = Kpoints3D().interpolated_points(atoms)  # _path

        # print ('kpath',kpoints)

        eigs = self.band_structure_eigs(kpath=kpoints, efermi=efermi).T

        for i, ii in enumerate(eigs):
            plt.plot(ii, color="b")

        plt.ylim(yrange)

        plt.savefig(filename)

        plt.close()

    def compare_dft_wann(
        self,
        vasprun_path="",
        energy_tol=0.75,
        plot=True,
        kp_labels_points=[],
        kp_labels=[],
        filename="compare.png",
    ):
        """Compare DFT and Wannier bands to check accuracy."""
        vrun = Vasprun(vasprun_path)
        kpoints = vrun.kpoints._kpoints
        fermi = vrun.efermi
        eigs_wan = self.band_structure_eigs(
            kpath=kpoints, efermi=vrun.efermi
        )  # -fermi#[::-1]#.T
        eigs_vrun = vrun.eigenvalues[0][:, :, 0] - fermi  # [::-1]#.T
        nbands = eigs_vrun.shape[1]
        nwann = eigs_wan.shape[1]
        info = {}
        print(
            "eigs.shape,eigs_vrun.shape",
            eigs_wan.shape,
            eigs_vrun.shape,
            nbands,
            nwann,
        )
        min_arr = []
        erange = [-energy_tol, energy_tol]
        for k in range(len(kpoints)):
            for n in eigs_wan[k]:
                diff_arr = []
                if n > erange[0] and n < erange[1]:
                    for v in eigs_vrun[k]:
                        diff = abs(n - v)
                        diff_arr.append(diff)
                if diff_arr != []:
                    tmp = np.min(diff_arr)
                    min_arr.append(tmp)
        maxdiff = "na"
        if min_arr != []:
            # print ('min_arr',min_arr)
            print("MAX diff", max(min_arr))
            maxdiff = max(min_arr)
        print("maxdiff", maxdiff)
        if plot:
            for i, ii in enumerate(eigs_wan.T):
                plt.plot(ii, color="b")
            for i, ii in enumerate(eigs_vrun.T):
                plt.plot(ii, color="r")
            plt.ylim([-energy_tol, energy_tol])
            if kp_labels_points != [] and kp_labels != []:
                plt.xticks(kp_labels_points, kp_labels)

            plt.savefig(filename)
            plt.close()
        info["eigs_wan"] = list(eigs_wan.tolist())
        info["eigs_vrun"] = list(eigs_vrun.tolist())
        info["kp_labels_points"] = list(kp_labels_points)
        info["kp_labels"] = kp_labels
        info["maxdiff"] = maxdiff
        info["efermi"] = fermi
        # print (info)
        return info  # ,eigs_wan.T,eigs_vrun.T

    def dos(
        self,
        kpoints=[],
        proj=None,
        efermi=0.0,
        xrange=None,
        nenergy=100,
        sig=0.02,
        pdf="dos.pdf",
        show=True,
    ):
        """Get density of states."""
        # plt.clf()

        # kgrid = generate_kgrid(grid)
        nk = len(kpoints)
        nwan = self.nwan

        vals = np.zeros((nk, nwan), dtype=float)
        pvals = np.zeros((nk, nwan), dtype=float)

        for i, k in enumerate(kpoints):
            val, vect, p = self.solve_ham(k, proj)
            vals[i, :] = val - efermi
            pvals[i, :] = p

        # print vals
        # print "pvals"
        # print pvals
        # print "np.sum pvals ", np.sum(np.sum(pvals))

        if xrange is None:
            vmin = np.min(vals[:])
            vmax = np.max(vals[:])
            vmin2 = vmin - (vmax - vmin) * 0.05
            vmax2 = vmax + (vmax - vmin) * 0.05
            xrange = [vmin2, vmax2]
            # plt.xlim(xrange)

        energies = np.arange(
            xrange[0],
            xrange[1] + 1e-5,
            (xrange[1] - xrange[0]) / float(nenergy),
        )
        dos = np.zeros(np.size(energies))
        pdos = np.zeros(np.size(energies))

        v = vals

        condmin = np.min(v[v > 0.0])
        valmax = np.max(v[v < 0.0])

        print("DOS BAND GAP ", condmin - valmax, "    ", valmax, " ", condmin)

        c = -0.5 / sig ** 2
        for i in range(np.size(energies)):
            arg = c * (v - energies[i]) ** 2
            dos[i] = np.sum(np.exp(arg))
            if proj is not None:
                pdos[i] = np.sum(np.exp(arg) * pvals)

        de = energies[1] - energies[0]
        dos = dos / sig / (2.0 * np.pi) ** 0.5 / float(nk)
        if proj is not None:
            pdos = pdos / sig / (2.0 * np.pi) ** 0.5 / float(nk)
        print("np.sum(dos) ", np.sum(dos * de))
        if proj is not None:
            print("np.sum(pdos) ", np.sum(pdos * de))

        return energies, dos, pdos

    def generate_supercell(
        self, supercell=[2, 2, 2], index=[0, 0, 0], sparse=False
    ):
        """Generate supercell."""
        t0 = time.time()

        nw = self.nwan

        factor = np.prod(supercell)
        NWAN = factor * self.nwan

        def plus_r(rold, subcell):
            rnew = subcell + rold
            # print ('cell',rnew,supercell)
            cellnew = rnew // supercell  # this is integer division
            subnew = rnew % supercell

            return cellnew, subnew

        def subcell_index(ss):
            t = (
                ss[0] * supercell[1] * supercell[2]
                + ss[1] * supercell[2]
                + ss[2]
            )
            return range(t * nw, (t + 1) * nw)

        RH_new = {}
        h_temp = np.zeros((self.nwan, self.nwan), dtype=complex)
        subcell = np.zeros(3, dtype=int)

        t1 = time.time()

        for ii in range(self.R.shape[0]):

            rold = np.array(self.R[ii, :], dtype=int)

            for i in range(supercell[0]):
                for j in range(supercell[1]):
                    for k in range(supercell[2]):
                        subcell[:] = [i, j, k]

                        cellnew, subnew = plus_r(rold, subcell)

                        if (
                            (index[0] > 0 and cellnew[0] != 0)
                            or (index[1] > 0 and cellnew[1] != 0)
                            or (index[2] > 0 and cellnew[2] != 0)
                        ):
                            continue

                        if tuple(cellnew) not in RH_new:
                            if sparse:
                                sps = scipy.sparse.csr_matrix  # TO CHECK
                                RH_new[tuple(cellnew)] = [
                                    cellnew,
                                    sps.lil_matrix(
                                        (NWAN, NWAN), dtype=complex
                                    ),
                                ]
                            else:
                                RH_new[tuple(cellnew)] = [
                                    cellnew,
                                    np.zeros((NWAN, NWAN), dtype=complex),
                                ]

                        h_temp[:, :] = np.reshape(
                            self.HR[ii, :], (self.nwan, self.nwan)
                        )

                        r1 = subcell_index(subcell)
                        r2 = subcell_index(subnew)

                        for c1, c2 in enumerate(r1):
                            for d1, d2 in enumerate(r2):
                                RH_new[tuple(cellnew)][1][c2, d2] += h_temp[
                                    c1, d1
                                ]

        t2 = time.time()

        nr = len(RH_new)
        hbig = WannierHam(nr=nr)
        # hbig = wan_ham()
        # if sparse:
        #    hbig.sparse = True

        # nwan = NWAN
        hbig.nwan = NWAN
        # hbig.nr = rn

        # R = np.zeros((nr, 3), dtype=float)
        hbig.R = np.zeros((nr, 3), dtype=float)
        if sparse:
            # HR = sps.lil_matrix((nr, NWAN ** 2), dtype=complex)
            hbig.HR = sps.lil_matrix((nr, NWAN ** 2), dtype=complex)
        else:
            # HR = np.zeros((nr, NWAN ** 2), dtype=complex)
            hbig.HR = np.zeros((nr, NWAN ** 2), dtype=complex)

        for c, i in enumerate(RH_new):
            h = RH_new[i][1]
            r = RH_new[i][0]
            if sparse:
                # HR[c, :] = sps.lil_matrix.reshape(h, NWAN * NWAN)
                hbig.HR[c, :] = sps.lil_matrix.reshape(h, NWAN * NWAN)
            else:
                # HR[c, :] = np.reshape(h, NWAN * NWAN)
                hbig.HR[c, :] = np.reshape(h, NWAN * NWAN)

            # R[c, :] = r
            hbig.R[c, :] = r

        t3 = time.time()
        print("TIME SUPERCELL", t1 - t0, t2 - t1, t3 - t2)
        return hbig

    def fermi_surf_2d(
        self,
        fermi=7.8163,
        origin=np.array([-1.0, -1.0, 0]),
        k1=np.array([4.0, 0.0, 0.0]),
        k2=np.array([0.0, 4.0, 0.0]),
        nk1=50,
        nk2=50,
        sig=0.3,
    ):
        """Generate 2D projected Fermi-surface."""
        K = np.zeros((nk1, nk2, 3), dtype=float)
        for c1 in range(nk1):
            for c2 in range(nk2):
                K[c1, c2, :] = (
                    origin
                    + k1 * float(c1) / float(nk1)
                    + k2 * float(c2) / float(nk2)
                )

        image = np.zeros((nk1, nk2))

        for c1 in range(nk1):
            for c2 in range(nk2):
                k = K[c1, c2, :]
                val, vect, p = self.solve_ham(k=k)
                image[c1, c2] = np.sum(
                    np.exp(-((val - fermi) ** 2) / sig ** 2)
                )

        return image


class Wannier90wout(object):
    """Construct wannier90.out related object."""

    def __init__(self, wout_path="wannier90.wout"):
        """Initialize with file path."""
        self.wout = wout_path

    def give_wannier_centers(self):
        """Get wannier charge centers."""
        f = open(self.wout, "r")
        lines = f.read().splitlines()
        f.close()
        final = False
        wan_cnts = []
        for ii, i in enumerate(lines):
            if "Final State" in i:
                final = True
            if final:
                if "WF centre and spread" in i:
                    tmp = [
                        float(j)
                        for j in i.split("(")[1].split(")")[0].split(",")
                    ]
                    wan_cnts.append(tmp)
        return wan_cnts


class Wannier90eig(object):
    """Construct wannier90.eig related object."""

    def __init__(self, weig_path="wannier90.eig"):
        """Initialize with file path."""
        self.weig = weig_path

    def give_wannier_eigs(self):
        """Get wannier eigs."""
        eigs = np.loadtxt(self.weig)
        return eigs

    def neigs(self):
        """Get wannier eigs."""
        ne = int(self.give_wannier_eigs()[-1][0])
        return ne

    def nk(self):
        """Get wannier eigs."""
        nk = int(len(self.give_wannier_eigs()) / self.neigs())
        return nk



from jarvis.core.pdb_atoms import read_pdb
import os

pdb_path = os.path.join(os.path.dirname(__file__), ".", "pdb101d.ent")


def test_pdb():
    pdb = read_pdb(pdb_path)
    # print (pdb.num_atoms,round(pdb.density,2))
    assert (pdb.num_atoms, round(pdb.density, 2)) == (448, 0.09)


# test_pdb()



"""Class for analyzing BoltzTrap outputs."""

import os
import numpy as np
from collections import OrderedDict, defaultdict

Ry_to_ev = 13.6056980659


class BoltzTrapOutput(object):
    """Analyze BoltzTrap output."""

    def __init__(
        self,
        path="",
        outtrans_data=[],
        intrans_data=[],
        condtens_fixdoping=[],
        halltens_fixdoping=[],
    ):
        """Specify boltztrap folder  as path to analze data."""
        self.path = path
        if outtrans_data == []:
            outtrans_data = self.read_outputtrans()
        self.outtrans_data = outtrans_data
        if intrans_data == []:
            intrans_data = self.read_intrans()
        self.intrans_data = intrans_data
        if condtens_fixdoping == []:
            condtens_fixdoping = self.read_condtens_fixdoping()
        self.condtens_fixdoping = condtens_fixdoping
        if halltens_fixdoping == []:
            halltens_fixdoping = self.read_halltens_fixdoping()
        self.halltens_fixdoping = halltens_fixdoping

    def to_dict(self):
        """Return output as a dictionary."""
        d = OrderedDict()
        d["path"] = self.path
        d["outtrans_data"] = self.outtrans_data
        d["intrans_data"] = self.intrans_data
        d["halltens_fixdoping"] = self.halltens_fixdoping
        d["condtens_fixdoping"] = self.condtens_fixdoping
        return d

    def read_intrans(self, filename=""):
        """Read intrans file."""
        if filename == "":
            filename = os.path.join(self.path, "boltztrap.intrans")
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        self.intrans_data = lines
        return lines

    def read_outputtrans(self, filename=""):
        """Read outtrans file."""
        gap = "na"
        Ef = "na"
        vbm = "na"
        cbm = "na"
        warning = "na"
        excessN_doping = "na"
        if filename == "":
            filename = os.path.join(self.path, "boltztrap.outputtrans")
        info = {}
        excessN_doping = {}
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        for ii, i in enumerate(lines):
            if "WARNING" in i:
                warning = i
            elif "Doping level number" in i:
                tmp_dop = float(i.split("=")[1].split()[0])
                tmp_excess = float(lines[ii + 1].split()[3])
                excessN_doping.setdefault(tmp_excess, tmp_dop)
            elif "Egap" in i:
                gap = float(i.split()[1]) * Ry_to_ev
            elif "VBM" in i:
                vbm = float(i.split()[1]) * Ry_to_ev
                cbm = float(i.split()[3]) * Ry_to_ev
                Ef = float(i.split()[5]) * Ry_to_ev

        info["gap"] = gap
        info["Ef"] = Ef
        info["vbm"] = vbm
        info["cbm"] = cbm
        info["warning"] = warning
        info["excessN_doping"] = excessN_doping
        # self.outtrans_data=info
        return info

    def dopinglevel_for_excessN(self, excessN):
        """Return doping level for excees concentration."""
        excessN_doping = self.read_outputtrans()["excessN_doping"]
        for i, j in excessN_doping.items():
            if np.isclose(i, excessN):
                return j

    def read_condtens_fixdoping(self, filename=""):
        """Read condtens_fixdoping file."""
        if filename == "":
            filename = os.path.join(self.path, "boltztrap.condtens_fixdoping")
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        full_doping_data = []
        for i in lines:
            if "#" not in i and len(i) > 2:
                full_doping_data.append([float(j) for j in i.split()])
        d = np.array(full_doping_data)
        all_data = {}
        p_dict = defaultdict(dict)
        n_dict = defaultdict(dict)
        for i in d:
            T = i[0]
            ef = i[29] * Ry_to_ev
            N = i[1]
            N_cm3 = self.dopinglevel_for_excessN(N)
            # print ('N,N_cm3',N,N_cm3)
            cond = i[2:11]
            seeb = i[11:20]
            kappa = i[20:29]
            if N > 0:
                info = {}
                info["cond"] = cond
                info["seeb"] = seeb
                info["kappa"] = kappa
                info["Ef"] = ef
                info["N_cm3"] = N_cm3
                p_dict[T][N] = info
            else:
                info = {}
                info["cond"] = cond
                info["seeb"] = seeb
                info["kappa"] = kappa
                info["Ef"] = ef
                info["N_cm3"] = N_cm3
                n_dict[T][N] = info
        all_data["p"] = p_dict
        all_data["n"] = n_dict
        # self.condtens_fixdoping=all_data
        return all_data

    def read_halltens_fixdoping(self, filename=""):
        """Read halltens file."""
        if filename == "":
            filename = os.path.join(self.path, "boltztrap.halltens_fixdoping")
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        full_doping_data = []
        for i in lines:
            if "#" not in i and len(i) > 2:
                full_doping_data.append([float(j) for j in i.split()])
        d = np.array(full_doping_data)
        all_data = {}
        p_dict = defaultdict(dict)
        n_dict = defaultdict(dict)
        for i in d:
            T = i[0]
            ef = i[29] * Ry_to_ev
            N = i[1]
            N_cm3 = self.dopinglevel_for_excessN(N)
            # print ('N,N_cm3',N,N_cm3)
            hall = i[2:29]
            if N > 0:
                info = {}
                info["hall"] = hall
                info["Ef"] = ef
                info["N_cm3"] = N_cm3
                p_dict[T][N] = info
            else:
                info = {}
                info["hall"] = hall
                info["Ef"] = ef
                info["N_cm3"] = N_cm3
                n_dict[T][N] = info
        all_data["p"] = p_dict
        all_data["n"] = n_dict
        # self.halltens_fixdoping=all_data
        return all_data

    @classmethod
    def from_dict(self, d={}):
        """Load from a dictionary."""
        return BoltzTrapOutput(
            path=d["path"],
            outtrans_data=d["outtrans_data"],
            intrans_data=d["intrans_data"],
            halltens_fixdoping=d["halltens_fixdoping"],
            condtens_fixdoping=d["condtens_fixdoping"],
        )


"""
if __name__ == "__main__":
    condtens_fix = "boltztrap/boltztrap.condtens_fixdoping"
    b = BoltzTrapOutput(
        path="boltztrap"
    ).read_condtens_fixdoping()
    b = BoltzTrapOutput(
        path="boltztrap"
    ).read_outputtrans()
    b = BoltzTrapOutput(
        path="boltztrap"
    ).to_dict()
    print(b)
"""



"""Helper functions for JSON files."""

import json


def loadjson(filename=""):
    """Provide helper function to load a json file."""
    f = open(filename, "r")
    d = json.load(f)
    f.close()
    return d


def dumpjson(data=[], filename=""):
    """Provide helper function to write a json file."""
    f = open(filename, "w")
    f.write(json.dumps(data))
    f.close()



"""Module to run phonopy. Assumes phonopy in PATH."""
from jarvis.io.vasp.outputs import Vasprun
import os
from jarvis.io.phonopy.inputs import PhonopyInputs


def run_phonopy(path="MAIN-ELAST"):
    """Run phonopy."""
    os.chdir(path)
    try:
        vrun = Vasprun(os.path.join(path, "vasprun.xml"))
        atoms = vrun.all_Structures[-1]
        PhonopyInputs(atoms=atoms).generate_all_files()
        cmd = "phonopy -p meshdos.conf"
        os.system(cmd)
        cmd = "phonopy -t meshdos.conf"
        os.system(cmd)
        cmd = "phonopy -p bandd.conf"
        os.system(cmd)
        cmd = "bandplot -o PBAND.png"
        os.system(cmd)
    except Exception:
        print("Cannot run phonopy. Check phonopy PATH and vasprun.")
        pass



from jarvis.core.kpoints import (
    Kpoints3D,
    generate_kpath,
    generate_kgrid,
    HighSymmetryKpoint3DFactory,
)
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from jarvis.core.atoms import Atoms
from jarvis.io.vasp.inputs import Poscar
import os
import tempfile


s1 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR"
    )
).atoms
s2 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR-Aem2"
    )
).atoms
s3 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR-C2m"
    )
).atoms
s4 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR-Cmcm"
    )
).atoms
s5 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR-P-1"
    )
).atoms
s6 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__),
        "..",
        "analysis",
        "structure",
        "POSCAR-tetragonal",
    )
).atoms
s7 = Poscar.from_file(
    os.path.join(
        os.path.dirname(__file__), "..", "analysis", "structure", "POSCAR-Pc"
    )
).atoms


def test_kp():
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    lattice_mat = Si.lattice_mat
    kp = Kpoints3D().automatic_length_mesh(lattice_mat=lattice_mat, length=20)
    td = kp.to_dict()
    fd = Kpoints3D.from_dict(td)
    sym = kp.high_symm_path(Si)._path
    x, y = kp.interpolated_points(Si)
    kpath = generate_kpath(kpath=[[0, 0, 0], [0, 0.5, 0.5]], num_k=5)
    kps = generate_kgrid(grid=[5, 5, 5])
    new_file, filename = tempfile.mkstemp()
    kp.write_file(filename)
    sym = kp.high_symm_path(s1)._path
    sym = kp.high_symm_path(s2)._path
    sym = kp.high_symm_path(s3)._path
    sym = kp.high_symm_path(s4)._path
    sym = kp.high_symm_path(s5)._path
    sym = kp.high_symm_path(s6)._path
    sym = kp.high_symm_path(s7)._path
    assert (len(x), x[0][0], y[0], kpath[0][0], kps[0][0]) == (
        166,
        0,
        "\Gamma",
        0.0,
        0.0,
    )
    kp = Kpoints3D().kpath(atoms=Si, unique_kp_only=True)


def test_extra_spgs():
    from jarvis.db.figshare import data

    d = data("dft_3d")
    few_spgs = {
        "JVASP-4663": 119,
        "JVASP-4666": 194,
        "JVASP-588": 160,
        "JVASP-4669": 191,
        "JVASP-4672": 141,
        "JVASP-581": 164,
        "JVASP-4687": 139,
        "JVASP-4693": 62,
        "JVASP-4696": 187,
        "JVASP-4711": 225,
        "JVASP-4714": 2,
        "JVASP-4723": 60,
        "JVASP-32": 167,
        "JVASP-107": 186,
        "JVASP-4756": 63,
        "JVASP-96": 216,
        "JVASP-4334": 123,
        "JVASP-4222": 11,
        "JVASP-4804": 156,
        "JVASP-329": 129,
        "JVASP-4852": 163,
        "JVASP-155": 61,
        "JVASP-4340": 51,
        "JVASP-4343": 15,
        "JVASP-4361": 14,
        "JVASP-137": 72,
        "JVASP-4879": 166,
        "JVASP-4885": 19,
        "JVASP-4894": 31,
        "JVASP-4216": 12,
        "JVASP-4918": 121,
        "JVASP-4948": 25,
        "JVASP-4957": 221,
        "JVASP-4960": 65,
        "JVASP-5059": 189,
        "JVASP-5071": 66,
        "JVASP-5074": 150,
        "JVASP-5086": 74,
        "JVASP-4388": 64,
        "JVASP-4391": 136,
        "JVASP-5155": 127,
        "JVASP-5185": 43,
        "JVASP-5197": 59,
        "JVASP-5212": 29,
        "JVASP-164": 176,
        "JVASP-5224": 137,
        "JVASP-5227": 148,
        "JVASP-4234": 193,
        "JVASP-5257": 7,
        "JVASP-5266": 140,
        "JVASP-4397": 33,
        "JVASP-5317": 8,
        "JVASP-5332": 13,
        "JVASP-5353": 16,
        "JVASP-5371": 4,
        "JVASP-5407": 229,
        "JVASP-5416": 147,
        "JVASP-5509": 52,
        "JVASP-5536": 70,
        "JVASP-5560": 71,
        "JVASP-5680": 28,
        "JVASP-5740": 6,
        "JVASP-5839": 162,
        "JVASP-5863": 79,
        "JVASP-110": 99,
        "JVASP-579": 38,
        "JVASP-4501": 5,
        "JVASP-91": 227,
        "JVASP-41": 154,
        "JVASP-4516": 96,
        "JVASP-4564": 82,
        "JVASP-4645": 130,
        "JVASP-152": 55,
        "JVASP-4792": 88,
        "JVASP-5041": 107,
        "JVASP-5425": 87,
        "JVASP-5464": 115,
        "JVASP-5650": 157,
        "JVASP-4450": 36,
        "JVASP-22520": 152,
        "JVASP-22523": 205,
        "JVASP-11998": 53,
        "JVASP-22528": 58,
        "JVASP-22533": 41,
        "JVASP-12091": 32,
        "JVASP-22541": 215,
        "JVASP-22543": 97,
        "JVASP-22549": 9,
        "JVASP-12103": 138,
        "JVASP-22575": 40,
        "JVASP-22602": 180,
        "JVASP-22611": 182,
        "JVASP-12022": 85,
        "JVASP-22637": 111,
        "JVASP-12139": 10,
        "JVASP-22709": 92,
        "JVASP-12060": 20,
        "JVASP-12064": 1,
        "JVASP-12194": 185,
        "JVASP-13885": 128,
        "JVASP-14096": 56,
        "JVASP-14020": 122,
        "JVASP-13904": 54,
        "JVASP-13783": 135,
        "JVASP-14213": 26,
        "JVASP-14034": 23,
        "JVASP-14158": 113,
        "JVASP-14256": 21,
        "JVASP-32150": 217,
        "JVASP-28392": 224,
        "JVASP-32180": 146,
        "JVASP-32197": 57,
        "JVASP-31813": 67,
        "JVASP-31819": 219,
        "JVASP-29262": 39,
        "JVASP-29281": 102,
        "JVASP-31825": 226,
        "JVASP-29425": 143,
        "JVASP-29441": 42,
        "JVASP-29520": 73,
        "JVASP-29526": 18,
        "JVASP-29555": 149,
        "JVASP-29597": 151,
        "JVASP-29705": 174,
        "JVASP-31921": 199,
        "JVASP-33147": 161,
        "JVASP-33220": 46,
        "JVASP-33344": 114,
        "JVASP-30263": 44,
        "JVASP-33832": 155,
        "JVASP-30458": 69,
        "JVASP-30461": 144,
        "JVASP-30518": 132,
        "JVASP-36540": 198,
        "JVASP-36548": 220,
        "JVASP-36568": 3,
        "JVASP-36573": 145,
        "JVASP-35061": 131,
        "JVASP-35137": 125,
        "JVASP-35222": 204,
        "JVASP-35344": 109,
        "JVASP-42053": 173,
        "JVASP-40216": 91,
        "JVASP-38594": 165,
        "JVASP-37330": 86,
        "JVASP-37573": 84,
        "JVASP-36714": 47,
        "JVASP-36754": 98,
        "JVASP-59313": 230,
        "JVASP-46893": 95,
        "JVASP-46896": 76,
        "JVASP-45779": 30,
        "JVASP-45831": 158,
        "JVASP-46446": 35,
        "JVASP-44393": 159,
        "JVASP-44773": 22,
        "JVASP-47741": 78,
        "JVASP-47811": 181,
        "JVASP-48055": 94,
        "JVASP-48916": 34,
        "JVASP-49907": 190,
        "JVASP-50342": 223,
        "JVASP-50360": 68,
        "JVASP-50431": 37,
        "JVASP-52902": 142,
        "JVASP-52377": 24,
        "JVASP-50791": 214,
        "JVASP-54512": 108,
        "JVASP-56567": 213,
        "JVASP-54867": 126,
        "JVASP-55180": 81,
        "JVASP-57780": 212,
        "JVASP-57807": 118,
        "JVASP-57816": 50,
        "JVASP-57100": 197,
        "JVASP-57138": 116,
        "JVASP-58233": 124,
        "JVASP-59682": 200,
        "JVASP-20180": 206,
        "JVASP-21448": 203,
        "JVASP-40468": 90,
        "JVASP-42914": 17,
        "JVASP-21594": 100,
        "JVASP-21706": 188,
        "JVASP-22783": 218,
        "JVASP-24006": 202,
        "JVASP-30879": 83,
        "JVASP-31186": 49,
        "JVASP-21031": 110,
        "JVASP-21116": 192,
        "JVASP-25245": 134,
        "JVASP-7066": 169,
        "JVASP-13017": 112,
        "JVASP-58953": 105,
        "JVASP-8682": 183,
        "JVASP-9902": 80,
        "JVASP-34882": 208,
        "JVASP-34330": 179,
        "JVASP-34502": 48,
        "JVASP-62982": 178,
    }
    """
 mem=[]
 spgs=[]
 pos=[]
 info=defaultdict()
 for i in d:
     a=Atoms.from_dict(i['atoms'])
     spg=Spacegroup3D(a).space_group_number
     if spg not in spgs:
        spgs.append(spg)
        pos.append(i['jid'])
        info[i['jid']]=spg
        print (info)
        if len(spgs)==230:
            break
 """
    for i, j in few_spgs.items():
        for ii in d:
            if ii["jid"] == i:

                a = Atoms.from_dict(ii["atoms"])
                spg = Spacegroup3D(a).space_group_number
                assert j == spg

                lattice_mat = a.lattice_mat
                kp = Kpoints3D().automatic_length_mesh(
                    lattice_mat=lattice_mat, length=40
                )
                sym = kp.high_symm_path(a)._path
                # print (ii['jid'],a)
                # TODO: following line failing for JVASP-4222
                # x, y = kp.interpolated_points(a)
                break


def test_highsym():
    kpts = HighSymmetryKpoint3DFactory().cubic()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().fcc()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().bcc()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().tet()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().bctet1(3, 2)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().bctet2(3, 2)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orc()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orcf1(1, 2, 3)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orcf2(1, 2, 3)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orcf3(1, 2, 3)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orci(1, 2, 3)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().orcc(1, 2, 3)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().hex()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().rhl1(47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().rhl2(47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mcl(2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mclc1(1, 2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mclc2(1, 2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mclc3(1, 2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mclc4(1, 2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().mclc5(1, 2, 3, 47)._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().tria()._kpoints["\\Gamma"]
    assert kpts[0] == 0
    kpts = HighSymmetryKpoint3DFactory().trib()._kpoints["\\Gamma"]
    assert kpts[0] == 0


test_extra_spgs()
# test_highsym()
# test_kp()



"""Modules for handling crystallographic Spacegroup related operations."""
from functools import reduce
from jarvis.core.lattice import Lattice
from jarvis.core.atoms import Atoms
import spglib
from jarvis.core.specie import Specie
import numpy as np
from numpy import sin, cos
import itertools
from fractions import gcd

# from numpy import gcd
# from math import gcd
import os


def unique_rows_2(a):
    """Remove duplicate rows."""
    order = np.lexsort(a.T)
    a = a[order]
    diff = np.diff(a, axis=0)
    ui = np.ones(len(a), "bool")
    ui[1:] = (diff != 0).any(axis=1)
    return a[ui]


def symmetrically_distinct_miller_indices(max_index=3, cvn_atoms=None):
    """Get unique miller indices for max_index."""
    # Need to work on this
    r1 = list(range(1, max_index + 1))
    r2 = list(range(-max_index, 1))
    r2.reverse()
    r3 = r1 + r2
    # r.reverse()
    r = r3
    # print ('sorted',r,sorted(r))
    conv_hkl_list = [
        miller
        for miller in itertools.product(r, r, r)
        if any([i != 0 for i in miller])
    ]
    spg = Spacegroup3D(cvn_atoms)._dataset
    rot = spg["rotations"]
    done = []
    unique_millers = []
    # print (conv_hkl_list)
    # print (sorted(conv_hkl_list, key=lambda x: x[0], reverse=True))
    for i in conv_hkl_list:
        d = abs(reduce(gcd, i))
        miller = tuple([int(k / d) for k in i])
        for j in rot:
            prod = list(np.dot(miller, j))
            if prod not in done:
                unique_millers.append(i)
                done.append(prod)
    uniq = unique_rows_2(np.array(unique_millers))
    return uniq


wyckoff_file = str(os.path.join(os.path.dirname(__file__), "Wyckoff.csv"))


def parse_wyckoff_csv(wyckoff_file):
    """Parse Wyckoff.csv from spglib.

    There are 530 data sets. For one example:
    9:C 1 2 1:::::::
    ::4:c:1:(x,y,z):(-x,y,-z)::
    ::2:b:2:(0,y,1/2):::
    ::2:a:2:(0,y,0):::
    """
    rowdata = []
    points = []
    hP_nums = [433, 436, 444, 450, 452, 458, 460]
    for i, line in enumerate(wyckoff_file):
        if line.strip() == "end of data":
            break
        rowdata.append(line.strip().split(":"))

        # 2:P -1  ::::::: <-- store line number if first element is number
        if rowdata[-1][0].isdigit():
            points.append(i)
    points.append(i)

    wyckoff = []
    for i in range(len(points) - 1):  # 0 to 529
        symbol = rowdata[points[i]][1]  # e.g. "C 1 2 1"
        if i + 1 in hP_nums:
            symbol = symbol.replace("R", "H", 1)
        wyckoff.append({"symbol": symbol.strip()})

    # When the number of positions is larger than 4,
    # the positions are written in the next line.
    # So those positions are connected.
    for i in range(len(points) - 1):
        count = 0
        wyckoff[i]["wyckoff"] = []
        for j in range(points[i] + 1, points[i + 1]):
            # Hook if the third element is a number (multiplicity), e.g.,
            #
            # 232:P 2/b 2/m 2/b:::::::  <- ignored
            # ::8:r:1:(x,y,z):(-x,y,-z):(x,-y+1/2,-z):(-x,-y+1/2,z)
            # :::::(-x,-y,-z):(x,-y,z):(-x,y+1/2,z):(x,y+1/2,-z)  <- ignored
            # ::4:q:..m:(x,0,z):(-x,0,-z):(x,1/2,-z):(-x,1/2,z)
            # ::4:p:..2:(0,y,1/2):(0,-y+1/2,1/2):(0,-y,1/2):(0,y+1/2,1/2)
            # ::4:o:..2:(1/2,y,0):(1/2,-y+1/2,0):(1/2,-y,0):(1/2,y+1/2,0)
            # ...
            if rowdata[j][2].isdigit():
                pos = []
                w = {
                    "letter": rowdata[j][3].strip(),
                    "multiplicity": int(rowdata[j][2]),
                    "site_symmetry": rowdata[j][4].strip(),
                    "positions": pos,
                }
                wyckoff[i]["wyckoff"].append(w)

                for k in range(4):
                    if rowdata[j][k + 5]:  # check if '(x,y,z)' or ''
                        count += 1
                        pos.append(rowdata[j][k + 5])
            else:
                for k in range(4):
                    if rowdata[j][k + 5]:
                        count += 1
                        pos.append(rowdata[j][k + 5])

        # assertion
        # for w in wyckoff[i]['wyckoff']:
        #    n_pos = len(w['positions'])
        #    n_pos *= len(lattice_symbols[wyckoff[i]['symbol'][0]])
        #    assert n_pos == w['multiplicity']

    return wyckoff


def read_wyckoff_csv(filename):
    """Read wyckoff_csv file."""
    with open(filename) as wyckoff_file:
        return parse_wyckoff_csv(wyckoff_file)


def get_wyckoff_position_operators(hall_number):
    """Get all Wyckoff operations for Hall number."""
    wyckoff = read_wyckoff_csv(wyckoff_file)
    operations = wyckoff[hall_number - 1]
    return operations


class Spacegroup3D(object):
    """
    Provide spacegroup related data for Atoms object.

    Currently uses spglib to derive spacegroup
    related information for 3D materials mainly
    """

    def __init__(self, atoms=[], dataset={}, symprec=1e-2, angle_tolerance=5):
        """
        Following information are needed for Spacegroup3D.

        If dataset is not provided, the default dataset is used.
        Args:
            atoms: jarvis.core.Atoms

            dataset: spacegroup dataset

            symprec: symmetry precision

            angle_tolerance: angle tolerance
        """
        self._dataset = dataset
        self._atoms = atoms
        self._symprec = symprec
        self._angle_tolerance = angle_tolerance
        if self._dataset == {}:
            spg = self.spacegroup_data()
            self._dataset = spg._dataset

    def spacegroup_data(self):
        """Provide spacegroup data from spglib."""
        phonopy_atoms = (
            self._atoms.lattice_mat,
            self._atoms.frac_coords,
            self._atoms.atomic_numbers,
        )
        dataset = spglib.get_symmetry_dataset(
            phonopy_atoms,
            symprec=self._symprec,
            angle_tolerance=self._angle_tolerance,
        )
        """
        keys = ('number',
        'hall_number',
        'international',
        'hall',
        'choice',
        'transformation_matrix',
        'origin_shift',
        'rotations',
        'translations',
        'wyckoffs',
        'site_symmetry_symbols',
        'equivalent_atoms',
        'mapping_to_primitive',
        'std_lattice',
        'std_types',
        'std_positions',
        'std_rotation_matrix',
        'std_mapping_to_primitive',
        'pointgroup')
        """
        return Spacegroup3D(
            atoms=self._atoms,
            symprec=self._symprec,
            angle_tolerance=self._angle_tolerance,
            dataset=dataset,
        )

    @property
    def space_group_symbol(self):
        """Get spacegroup symbol."""
        # spg = self.spacegroup_data()
        return self._dataset["international"]

    @property
    def space_group_number(self):
        """Get spacegroup number."""
        # spg = self.spacegroup_data()
        return self._dataset["number"]

    @property
    def primitive_atoms(self):
        """Get primitive atoms."""
        phonopy_atoms = (
            self._atoms.lattice_mat,
            self._atoms.frac_coords,
            self._atoms.atomic_numbers,
        )
        lattice, scaled_positions, numbers = spglib.find_primitive(
            phonopy_atoms, symprec=self._symprec
        )
        elements = self._atoms.elements
        el_dict = {}
        for i in elements:
            el_dict.setdefault(Specie(i).Z, i)
        prim_elements = [el_dict[i] for i in numbers]
        prim_atoms = Atoms(
            lattice_mat=lattice,
            elements=prim_elements,
            coords=scaled_positions,
            cartesian=False,
        )
        return prim_atoms

    @property
    def refined_atoms(self):
        """Refine atoms based on spacegroup data."""
        phonopy_atoms = (
            self._atoms.lattice_mat,
            self._atoms.frac_coords,
            self._atoms.atomic_numbers,
        )
        lattice, scaled_positions, numbers = spglib.refine_cell(
            phonopy_atoms, self._symprec, self._angle_tolerance
        )
        elements = self._atoms.elements
        el_dict = {}
        for i in elements:
            el_dict.setdefault(Specie(i).Z, i)
        ref_elements = [el_dict[i] for i in numbers]
        ref_atoms = Atoms(
            lattice_mat=lattice,
            elements=ref_elements,
            coords=scaled_positions,
            cartesian=False,
        )
        return ref_atoms

    @property
    def crystal_system(self):
        """Get crystal system."""
        n = self._dataset["number"]

        def f(i, j):
            return i <= n <= j

        cs = {
            "triclinic": (1, 2),
            "monoclinic": (3, 15),
            "orthorhombic": (16, 74),
            "tetragonal": (75, 142),
            "trigonal": (143, 167),
            "hexagonal": (168, 194),
            "cubic": (195, 230),
        }

        crystal_sytem = None

        for k, v in cs.items():
            if f(*v):
                crystal_sytem = k
                break
        return crystal_sytem

    @property
    def lattice_system(self):
        """Get lattice system."""
        n = self._dataset["number"]
        system = self.crystal_system
        if n in [146, 148, 155, 160, 161, 166, 167]:
            return "rhombohedral"
        elif system == "trigonal":
            return "hexagonal"
        else:
            return system

    @property
    def point_group_symbol(self):
        """Get pointgroup."""
        return self._dataset["pointgroup"]

    @property
    def conventional_standard_structure(
        self, tol=1e-5, international_monoclinic=True
    ):
        """
        Give a conventional cell according to certain conventions.

        The conventionss are defined in Setyawan, W., & Curtarolo,
        S. (2010). High-throughput electronic band structure calculations:
        Challenges and tools. Computational Materials Science,
        49(2), 299-312. doi:10.1016/j.commatsci.2010.05.010
        They basically enforce as much as possible
        norm(a1)<norm(a2)<norm(a3)
        Returns:
            The structure in a conventional standardized cell
        """
        struct = self.refined_atoms
        latt = struct.lattice
        latt_type = self.lattice_system
        sorted_lengths = sorted(latt.abc)
        sorted_dic = sorted(
            [
                {"vec": latt.matrix[i], "length": latt.abc[i], "orig_index": i}
                for i in [0, 1, 2]
            ],
            key=lambda k: k["length"],
        )

        if latt_type in ("orthorhombic", "cubic"):
            # you want to keep the c axis where it is
            # to keep the C- settings
            transf = np.zeros(shape=(3, 3))
            if self.space_group_symbol.startswith("C"):
                transf[2] = [0, 0, 1]
                a, b = sorted(latt.abc[:2])
                sorted_dic = sorted(
                    [
                        {
                            "vec": latt.matrix[i],
                            "length": latt.abc[i],
                            "orig_index": i,
                        }
                        for i in [0, 1]
                    ],
                    key=lambda k: k["length"],
                )
                for i in range(2):
                    transf[i][sorted_dic[i]["orig_index"]] = 1
                c = latt.abc[2]
            elif self.space_group_symbol.startswith(
                "A"
            ):  # change to C-centering to match Setyawan/Curtarolo convention
                transf[2] = [1, 0, 0]
                a, b = sorted(latt.abc[1:])
                sorted_dic = sorted(
                    [
                        {
                            "vec": latt.matrix[i],
                            "length": latt.abc[i],
                            "orig_index": i,
                        }
                        for i in [1, 2]
                    ],
                    key=lambda k: k["length"],
                )
                for i in range(2):
                    transf[i][sorted_dic[i]["orig_index"]] = 1
                c = latt.abc[0]
            else:
                for i in range(len(sorted_dic)):
                    transf[i][sorted_dic[i]["orig_index"]] = 1
                a, b, c = sorted_lengths
            latt = Lattice.orthorhombic(a, b, c)

        elif latt_type == "tetragonal":
            # find the "a" vectors
            # it is basically the vector repeated two times
            transf = np.zeros(shape=(3, 3))
            a, b, c = sorted_lengths
            for d in range(len(sorted_dic)):
                transf[d][sorted_dic[d]["orig_index"]] = 1

            if abs(b - c) < tol and abs(a - c) > tol:
                a, c = c, a
                transf = np.dot([[0, 0, 1], [0, 1, 0], [1, 0, 0]], transf)
            latt = Lattice.tetragonal(a, c)
        elif latt_type in ("hexagonal", "rhombohedral"):
            # for the conventional cell representation,
            # we allways show the rhombohedral lattices as hexagonal

            # check first if we have the refined structure shows a rhombohedral
            # cell
            # if so, make a supercell
            a, b, c = latt.abc
            if np.all(np.abs([a - b, c - b, a - c]) < 0.001):
                struct.make_supercell(((1, -1, 0), (0, 1, -1), (1, 1, 1)))
                a, b, c = sorted(struct.lattice.abc)

            if abs(b - c) < 0.001:
                a, c = c, a
            new_matrix = [
                [a / 2, -a * np.sqrt(3) / 2, 0],
                [a / 2, a * np.sqrt(3) / 2, 0],
                [0, 0, c],
            ]
            latt = Lattice(new_matrix)
            transf = np.eye(3, 3)

        elif latt_type == "monoclinic":
            # You want to keep the c axis where it is to keep the C- settings

            if self.space_group_symbol.startswith("C"):
                transf = np.zeros(shape=(3, 3))
                transf[2] = [0, 0, 1]
                sorted_dic = sorted(
                    [
                        {
                            "vec": latt.matrix[i],
                            "length": latt.abc[i],
                            "orig_index": i,
                        }
                        for i in [0, 1]
                    ],
                    key=lambda k: k["length"],
                )
                a = sorted_dic[0]["length"]
                b = sorted_dic[1]["length"]
                c = latt.abc[2]
                new_matrix = None
                for t in itertools.permutations(list(range(2)), 2):
                    m = latt.matrix
                    latt2 = Lattice([m[t[0]], m[t[1]], m[2]])
                    lengths = latt2.abc
                    angles = latt2.angles
                    if angles[0] > 90:
                        # if the angle is > 90 we invert a and b to get
                        # an angle < 90
                        a, b, c, alpha, beta, gamma = Lattice(
                            [-m[t[0]], -m[t[1]], m[2]]
                        ).parameters
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = -1
                        transf[1][t[1]] = -1
                        transf[2][2] = 1
                        alpha = np.pi * alpha / 180
                        new_matrix = [
                            [a, 0, 0],
                            [0, b, 0],
                            [0, c * cos(alpha), c * sin(alpha)],
                        ]
                        continue

                    elif angles[0] < 90:
                        transf = np.zeros(shape=(3, 3))
                        # print ('464-470')
                        transf[0][t[0]] = 1
                        transf[1][t[1]] = 1
                        transf[2][2] = 1
                        a, b, c = lengths
                        alpha = np.pi * angles[0] / 180
                        new_matrix = [
                            [a, 0, 0],
                            [0, b, 0],
                            [0, c * cos(alpha), c * sin(alpha)],
                        ]

                if new_matrix is None:
                    # print ('479-482')
                    # this if is to treat the case
                    # where alpha==90 (but we still have a monoclinic sg
                    new_matrix = [[a, 0, 0], [0, b, 0], [0, 0, c]]
                    transf = np.zeros(shape=(3, 3))
                    for c in range(len(sorted_dic)):
                        transf[c][sorted_dic[c]["orig_index"]] = 1
            # if not C-setting
            else:
                # try all permutations of the axis
                # keep the ones with the non-90 angle=alpha
                # and b<c
                new_matrix = None
                for t in itertools.permutations(list(range(3)), 3):
                    m = latt.matrix
                    a, b, c, alpha, beta, gamma = Lattice(
                        [m[t[0]], m[t[1]], m[t[2]]]
                    ).parameters
                    if alpha > 90 and b < c:
                        a, b, c, alpha, beta, gamma = Lattice(
                            [-m[t[0]], -m[t[1]], m[t[2]]]
                        ).parameters
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = -1
                        transf[1][t[1]] = -1
                        transf[2][t[2]] = 1
                        alpha = np.pi * alpha / 180
                        new_matrix = [
                            [a, 0, 0],
                            [0, b, 0],
                            [0, c * cos(alpha), c * sin(alpha)],
                        ]
                        continue
                    elif alpha < 90 and b < c:
                        # print ('510-515')
                        transf = np.zeros(shape=(3, 3))
                        transf[0][t[0]] = 1
                        transf[1][t[1]] = 1
                        transf[2][t[2]] = 1
                        alpha = np.pi * alpha / 180
                        new_matrix = [
                            [a, 0, 0],
                            [0, b, 0],
                            [0, c * cos(alpha), c * sin(alpha)],
                        ]
                if new_matrix is None:
                    # print ('523-530')
                    # this if is to treat the case
                    # where alpha==90 (but we still have a monoclinic sg
                    new_matrix = [
                        [sorted_lengths[0], 0, 0],
                        [0, sorted_lengths[1], 0],
                        [0, 0, sorted_lengths[2]],
                    ]
                    transf = np.zeros(shape=(3, 3))
                    for c in range(len(sorted_dic)):
                        transf[c][sorted_dic[c]["orig_index"]] = 1

            if international_monoclinic:
                # The above code makes alpha the non-right angle.
                # The following will convert to proper international convention
                # that beta is the non-right angle.
                op = [[0, 1, 0], [1, 0, 0], [0, 0, -1]]
                transf = np.dot(op, transf)
                new_matrix = np.dot(op, new_matrix)
                beta = Lattice(new_matrix).beta
                if beta < 90:
                    op = [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
                    transf = np.dot(op, transf)
                    new_matrix = np.dot(op, new_matrix)

            latt = Lattice(new_matrix)

        elif latt_type == "triclinic":
            # we use a LLL Minkowski-like reduction for the triclinic cells
            struct = struct.get_lll_reduced_structure()

            a, b, c = latt.abc  # lengths
            alpha, beta, gamma = [np.pi * i / 180 for i in latt.angles]
            new_matrix = None
            test_matrix = [
                [a, 0, 0],
                [b * cos(gamma), b * sin(gamma), 0.0],
                [
                    c * cos(beta),
                    c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma),
                    c
                    * np.sqrt(
                        sin(gamma) ** 2
                        - cos(alpha) ** 2
                        - cos(beta) ** 2
                        + 2 * cos(alpha) * cos(beta) * cos(gamma)
                    )
                    / sin(gamma),
                ],
            ]

            def is_all_acute_or_obtuse(m):
                recp_angles = np.array(Lattice(m).reciprocal_lattice().angles)
                return np.all(recp_angles <= 90) or np.all(recp_angles > 90)

            if is_all_acute_or_obtuse(test_matrix):
                transf = np.eye(3)
                new_matrix = test_matrix

            test_matrix = [
                [-a, 0, 0],
                [b * cos(gamma), b * sin(gamma), 0.0],
                [
                    -c * cos(beta),
                    -c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma),
                    -c
                    * np.sqrt(
                        sin(gamma) ** 2
                        - cos(alpha) ** 2
                        - cos(beta) ** 2
                        + 2 * cos(alpha) * cos(beta) * cos(gamma)
                    )
                    / sin(gamma),
                ],
            ]

            if is_all_acute_or_obtuse(test_matrix):
                transf = [[-1, 0, 0], [0, 1, 0], [0, 0, -1]]
                new_matrix = test_matrix

            test_matrix = [
                [-a, 0, 0],
                [-b * cos(gamma), -b * sin(gamma), 0.0],
                [
                    c * cos(beta),
                    c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma),
                    c
                    * np.sqrt(
                        sin(gamma) ** 2
                        - cos(alpha) ** 2
                        - cos(beta) ** 2
                        + 2 * cos(alpha) * cos(beta) * cos(gamma)
                    )
                    / sin(gamma),
                ],
            ]

            if is_all_acute_or_obtuse(test_matrix):
                transf = [[-1, 0, 0], [0, -1, 0], [0, 0, 1]]
                new_matrix = test_matrix

            test_matrix = [
                [a, 0, 0],
                [-b * cos(gamma), -b * sin(gamma), 0.0],
                [
                    -c * cos(beta),
                    -c * (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma),
                    -c
                    * np.sqrt(
                        sin(gamma) ** 2
                        - cos(alpha) ** 2
                        - cos(beta) ** 2
                        + 2 * cos(alpha) * cos(beta) * cos(gamma)
                    )
                    / sin(gamma),
                ],
            ]
            if is_all_acute_or_obtuse(test_matrix):
                transf = [[1, 0, 0], [0, -1, 0], [0, 0, -1]]
                new_matrix = test_matrix

            latt = Lattice(new_matrix)

        new_coords = np.dot(transf, np.transpose(struct.frac_coords)).T
        new_struct = Atoms(
            lattice_mat=latt.matrix,
            elements=struct.elements,
            coords=new_coords,
            cartesian=False,
        )
        return new_struct


def parse_xyz_string(xyz_string):
    """
    Convert xyz info to translation and rotation vectors.

    Adapted from pymatgen.
    Args:
        xyz_string: string of the form 'x, y, z', '-x, -y, z',
            '-2y+1/2, 3x+1/2, z-y+1/2', etc.
    Returns:
        translation and rotation vectors.
    """
    from jarvis.core.utils import parse_xyz_string

    return parse_xyz_string(xyz_string)


def operate_affine(cart_coord=[], affine_matrix=[]):
    """Operate affine method."""
    affine_point = np.array([cart_coord[0], cart_coord[1], cart_coord[2], 1])
    return np.dot(np.array(affine_matrix), affine_point)[0:3]


def get_new_coord_for_xyz_sym(frac_coord=[], xyz_string=""):
    """Obtain new coord from xyz string."""
    from jarvis.core.utils import get_new_coord_for_xyz_sym

    return get_new_coord_for_xyz_sym(
        frac_coord=frac_coord, xyz_string=xyz_string
    )


def check_duplicate_coords(coords=[], coord=[]):
    """Check if a coordinate exists."""
    from jarvis.core.utils import check_duplicate_coords

    return check_duplicate_coords(coords=coords, coord=coord)


"""
if __name__ == "__main__":
    x = get_wyckoff_position_operators(488)
    print (x)
    import sys
    #sys.exit()
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    spg = Spacegroup3D(atoms=Si)  # .spacegroup_data()
    #print(spg.space_group_symbol)
    #print(spg.space_group_number)
    #primt = spg.primitive_atoms
    #print("primt", primt)
    #print("cryst_sys", spg.crystal_system)
    #print("point group", spg.point_group_symbol)
    #print("cvn", spg.conventional_standard_structure)
    cvn = spg.conventional_standard_structure
    ml=symmetrically_distinct_miller_indices(max_index=3, cvn_atoms=cvn)
    print ('miller=',ml,len(ml))
#    ml_pmg=get_symmetrically_distinct_miller_indices(cvn.pymatgen_converter(),max_index=3)
#    print ('millerpmg=',ml_pmg,len(ml_pmg))
    # pmg=Si.pymatgen_converter()
    # from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
    # spg=SpacegroupAnalyzer(pmg)
    # print (spg.get_space_group_symbol(),spg.get_space_group_number())
    # print (pmg.get_primitive_structure())
"""



from jarvis.analysis.stm.tersoff_hamann import TersoffHamannSTM
import matplotlib.pyplot as plt
import os

name = os.path.join(os.path.dirname(__file__), "PARCHG")
from jarvis.core.image import Image
from io import BytesIO


def test_th_stm():
    plt.switch_backend("agg")
    TH_STM1 = TersoffHamannSTM(chg_name=name)
    byte_io = BytesIO()
    t1 = TH_STM1.constant_height(filename=byte_io)

    t1 = TH_STM1.constant_height(filename="test.png")
    im = Image.from_file("test.png")
    plt.imshow(
        im.fourier_transform2D(use_crop=True, zoom_factor=50)
        .rotate(angle=0)
        .black_and_white(threshold=0.05)
        .values,
        cmap="Greys",
    )
    p = byte_io.getvalue()  # .decode('UTF-8')
    # print ('p',p)
    t1 = TH_STM1.constant_height()
    TH_STM2 = TersoffHamannSTM(chg_name=name)
    t2 = TH_STM2.constant_current()
    t2 = TH_STM2.constant_current(pc=5)
    cmd = "rm *.png"
    os.system(cmd)


# test_th_stm()



"""
Code to predict properties and their uncertainty.

ML model used: lgbm
"""

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
import pickle
from jarvis.ai.pkgs.utils import regr_scores
from collections import OrderedDict
from jarvis.ai.pkgs.lgbm.regression import get_lgbm


def quantile_regr_predint(
    x,
    y,
    jid,
    cv=2,
    n_jobs=-1,
    n_iter=10,
    random_state=508842607,
    scoring="neg_mean_absolute_error",
    prop="exfoliation_energy",
    write_files=True,
):
    """
    Perform Quantile regression and determine prediction intervals.

    LOWER_ALPHA = 0.16
    Mid model uses ls as loss function, not quantile, to
        optimize for the mean, not the median
    UPPER_ALPHA = 0.84
    This choice of LOWER_ALPHA, UPPER_ALPHA gives a prediction
    interval ideally equal to 0.68, i.e. 1 standard deviation.
    However, the number of in-bound prediction must be computed
    for the specific fitted models, and that gives the true meaning
    of the uncertainties computed here.
    See:
    https://machinelearningmastery.com/prediction-intervals-for-machine-learning
    https://www.inovex.de/blog/uncertainty-quantification-deep-learning
    """
    # TODO: Make writing file in proper python format

    # STEP-2: Splitting the data
    # ***************************
    # 90-10% split for train test

    X_train, X_test, y_train, y_test, jid_train, jid_test = train_test_split(
        x, y, jid, random_state=1, test_size=0.1
    )
    # print ('lenx len y',len(x[0]),len(y))

    # STEP-3: Use a specific ML model
    # ********************************

    # Set lower and upper quantile
    # StanDev
    LOWER_ALPHA = 0.16
    # MID_ALPHA = 0.50
    UPPER_ALPHA = 0.84

    # LOWER Model
    # ===========
    scaler = StandardScaler().fit(X_train)
    scaler.transform(X_train)
    scaler.transform(X_test)

    objective = "quantile"
    alpha = LOWER_ALPHA
    print("Prima di lgbm for LOWER model")
    lower_model = get_lgbm(
        X_train,
        X_test,
        y_train,
        y_test,
        cv,
        n_jobs,
        scoring,
        n_iter,
        objective,
        alpha,
        random_state,
    )
    print("Dopo lgbm for LOWER model")
    name = str(prop) + str("_lower")
    filename = str("pickle2-") + str(name) + str(".pk")
    pickle.dump(lower_model, open(filename, "wb"))

    # MID Model
    # =========
    scaler = StandardScaler().fit(X_train)
    scaler.transform(X_train)
    scaler.transform(X_test)

    # mid_model.fit(X_train, y_train)
    objective = "regression"
    alpha = 0.9
    print("Prima di lgbm for MID model")
    mid_model = get_lgbm(
        X_train,
        X_test,
        y_train,
        y_test,
        cv,
        n_jobs,
        scoring,
        n_iter,
        objective,
        alpha,
        random_state,
    )
    print("Dopo lgbm for MID model")
    name = str(prop) + str("_mid")
    filename = str("pickle2-") + str(name) + str(".pk")
    pickle.dump(mid_model, open(filename, "wb"))

    # UPPER Model
    # ===========
    scaler = StandardScaler().fit(X_train)
    scaler.transform(X_train)
    scaler.transform(X_test)

    # upper_model.fit(X_train, y_train)
    objective = "quantile"
    alpha = UPPER_ALPHA
    print("Prima di lgbm for UPPER model")
    upper_model = get_lgbm(
        X_train,
        X_test,
        y_train,
        y_test,
        cv,
        n_jobs,
        scoring,
        n_iter,
        objective,
        alpha,
        random_state,
    )
    print("Dopo lgbm for UPPER model")
    name = str(prop) + str("_upper")
    filename = str("pickle2-") + str(name) + str(".pk")
    pickle.dump(upper_model, open(filename, "wb"))

    # PREDICTIONS and UQ
    lower = lower_model.predict(X_test)
    mid = mid_model.predict(X_test)
    upper = upper_model.predict(X_test)
    actual = y_test

    print("Model       mae     rmse")
    reg_sc = regr_scores(y_test, lower)
    info = OrderedDict()
    info["MAE_Lower"] = reg_sc["mae"]
    print("Lower:", round(reg_sc["mae"], 3), round(reg_sc["rmse"], 3))
    reg_sc = regr_scores(y_test, mid)
    info["MAE_Mid"] = reg_sc["mae"]
    print("Mid:", round(reg_sc["mae"], 3), round(reg_sc["rmse"], 3))
    reg_sc = regr_scores(y_test, upper)
    info["MAE_Upper"] = reg_sc["mae"]
    print("Upper:", round(reg_sc["mae"], 3), round(reg_sc["rmse"], 3))

    # Calculate the absolute error associated with prediction intervals
    # in_bounds = actual.between(left=lower, right=upper)
    if write_files:
        fout1 = open("Intervals.dat", "w")
        fout2 = open("Intervals1.dat", "w")
        line0 = "#    Jid      Observed       pred_Lower"
        line1 = "       pred_Mid        pred_Upper\n"
        line = line0 + line1
        fout1.write(line)
        line0 = "#    Jid      Observed       pred_Lower    AbsErr(Lower)"
        line1 = "     pred_Mid    AbsErr(Mid)     pred_Upper"
        line2 = "     AbsErr(Upper)    AbsErrInterval    Pred_inBounds\n"
        line = line0 + line1 + line2
        fout2.write(line)
        sum = 0.0
        count = 0
        MAE_err = 0.0
        for ii in range(len(actual)):
            true = float(actual[ii])
            llow = float(lower[ii])
            mmid = float(mid[ii])
            uupper = float(upper[ii])
            err = abs((uupper - llow) * 0.5)
            diff = true - mmid
            real_err = abs(diff)
            err_err = abs(real_err - err)
            MAE_err = MAE_err + err_err
            if abs(diff) < err:
                count = count + 1
                in_bounds = "True"
            else:
                in_bounds = "False"
            absolute_error_lower = abs(lower[ii] - actual[ii])
            absolute_error_mid = abs(mid[ii] - actual[ii])
            absolute_error_upper = abs(upper[ii] - actual[ii])
            absolute_error_interval = (
                absolute_error_lower + absolute_error_upper
            ) / 2.0

            line = (
                str(ii)
                + " "
                + jid[ii]
                + " "
                + str(actual[ii])
                + " "
                + str(lower[ii])
                + " "
                + str(mid[ii])
                + " "
                + str(upper[ii])
                + "\n"
            )
            sum = sum + float(absolute_error_interval)
            line2 = (
                str(ii)
                + " "
                + jid[ii]
                + " "
                + str(actual[ii])
                + " "
                + str(lower[ii])
                + " "
                + str(absolute_error_lower)
                + " "
                + str(mid[ii])
                + " "
                + str(absolute_error_mid)
                + " "
                + str(upper[ii])
                + " "
                + str(absolute_error_upper)
                + " "
                + str(absolute_error_interval)
                + " "
                + str(in_bounds)
                + "\n"
            )
            fout1.write(line)
            fout2.write(line2)
    print("")
    print("Number of test materials= " + str(len(actual)))
    print(
        "Percentage of in-bound results= "
        + str((float(count) / (len(actual))) * 100)
        + "%"
    )
    print(" ")
    MAE_error = float(MAE_err) / (len(actual))
    print("MAE predicted error (err=0.5*(High-Low))= " + str(MAE_error))
    info["MAE_Error"] = MAE_error
    return info



"""Module to prepare input files for WIEN2k."""
from scipy import array, zeros, sqrt, dot
import os
from fractions import gcd
import numpy as np
from functools import reduce
# from jarvis.core.atoms import Atoms
from jarvis.core.kpoints import Kpoints3D


def get_wien_kpoints(
    atoms=None, write_file=False, line_density=1, filename="MyKpoints"
):
    """Get WIEN2k style kpoints for bandstructure calculation."""
    uniqe_lbl = []
    uniqe_k = []
    kpoints, labels = Kpoints3D().interpolated_points(
        atoms, line_density=line_density
    )
    kp = Kpoints3D().high_kpath(atoms)["kpoints"]
    for i, j in kp.items():
        uniqe_lbl.append(i)
        uniqe_k.append(j)
    legend = uniqe_lbl
    BS = uniqe_k
    data = []
    BS = array(BS, dtype=float)
    Nt = 500
    dl = zeros(len(BS) - 1)
    for i in range(len(BS) - 1):
        dr = BS[i + 1] - BS[i]
        dl[i] = sqrt(dot(dr, dr))
    dN = dl / sum(dl) * Nt
    Ni = [int(round(dn)) for dn in dN]

    Mi = zeros(len(Ni), dtype=int)
    for i in range(len(Ni)):
        tmp = np.concatenate((BS[i + 1], BS[i]))
        fracts = 1 // array(list(filter(lambda x: x > 1e-6, tmp)))
        fact = int(reduce(lambda x, y: x * y // gcd(x, y), fracts))
        if Ni[i] % fact == 0:
            Mi[i] = Ni[i]
        else:
            Mi[i] = Ni[i] * fact

    for p in range(len(Ni)):
        NAME = legend[p]
        for i in range(Ni[p]):
            kint = Mi[p] * BS[p] + (BS[p + 1] - BS[p]) * i * Mi[p] // Ni[p]
            if i > 0:
                NAME = "   "
            data.append([NAME, kint[0], kint[1], kint[2], Mi[p], 1.0])
    NAME = legend[-1]
    kint = BS[-1] * Mi[-1]
    data.append([NAME, kint[0], kint[1], kint[2], Mi[-1], 1.0])

    if write_file:
        f = open(filename, "w")
        for i in data:
            line = (
                str(i[0])
                + str(" ")
                + str(i[1])
                + str(" ")
                + str(i[2])
                + str(" ")
                + str(i[3])
                + str(" ")
                + str(i[4])
                + str(" ")
                + str(i[5])
                + str("\n")
            )
            f.write(line)
        line = str("END \n")
        f.write(line)
        f.close()
    return data


def prepare_wien_input(atoms=None, filename="case.cif"):
    """Prepare basic WIEN2k input files."""
    atoms.write_cif(filename=filename)
    cmd = str("cif2struct ") + str(filename)
    os.system(cmd)
    cmd = str(
        "init_lapw -b -red 0 -vxc 13 -ecut -7.0  -numk 100 -sp >init_w2k_out"
    )
    os.system(cmd)
    cmd = str("runsp_lapw -cc 0.0001 -ec 0.0001 -p -i 500  >run_w2k_out")
    os.system(cmd)


"""
if __name__ == "__main__":
    legend = ["Gamma", "H", "P", "N", "Gamma", "P"]
    BS = [
        [0, 0, 0],
        [0, 1, 0],
        [1 / 2.0, 1 / 2.0, 1 / 2.0],
        [1 / 2.0, 1 / 2.0, 0],
        [0, 0, 0],
        [1 / 2.0, 1 / 2.0, 1 / 2.0],
    ]

    s = Atoms.from_poscar("POSCAR")

    data = get_wien_kpoints(atoms=s, write_file=True)

    for i in data:
        print("%-10s%5d%5d%5d%5d%5.1f" % (i[0], i[1], i[2], i[3], i[4], i[5]))
"""



"""
Code to calculate spin-orbit spillage.

Reference:
https://www.nature.com/articles/s41598-019-45028-y
https://www.nature.com/articles/s41524-020-0319-4
https://arxiv.org/abs/2102.00237
"""

import numpy as np
from jarvis.io.vasp.outputs import Wavecar  # vaspwfc


class Spillage(object):
    """
    Spin-orbit spillage criteria.

    Predict whether a material is topologically non-trival.
    The spillage criteria physically signifies
    number of band-inverted electrons.
    A non-zero, high value (generally >0.5) suggests non-trivial behavior
    """

    def __init__(self, wf_noso="", wf_so=""):
        """
        Require path to WAVECAR files with and without LSORBIT = .TRUE.

        Args:
            wf_noso : WAVECAR without spin-orbit coupling

            wf_so : WAVECAR with spin-orbit coupling
        """
        self.wf_noso = wf_noso
        self.wf_so = wf_so

    def isclose(self, n1, n2, rel_tol=1e-7):
        """Check if n1 and n2 are close."""
        if abs(n1 - n2) < rel_tol:
            return True
        else:
            return False

    def orth(self, A):
        """Orthogonalize a vector."""
        # As we do not store overlap matrix, we orthogonalize the
        # Wavefunctions.
        u, s, vh = np.linalg.svd(A, full_matrices=False)
        M, N = A.shape
        eps = np.finfo(float).eps
        tol = max(M, N) * np.amax(s) * eps
        num = np.sum(s > tol, dtype=int)
        Q = u[:, :num]
        return Q, num

    def overlap_so_spinpol(self):
        """Calculate spillage."""
        noso = Wavecar(filename=self.wf_noso, lsorbit=False)
        so = Wavecar(filename=self.wf_so, lsorbit=True)
        # band gap stuff. not needed per se, just a useful sanity check
        noso_k, noso_bands = noso.readWFBand()
        so_k, so_bands = so.readWFBand()

        # Calculate the number of occupied bands in non-soc calculation
        # at each k-point

        # Start from deep energy levels and then finds
        # first unoccupied bands with occupation less than 50%
        # noso._kvecs[nk1 - 1, :]: number of kpoints are rows,
        # number of columns three, kpoint units

        nelec_list = []
        for nk1 in range(1, noso._nkpts + 1):  # no spin orbit kpoints loop
            knoso = noso._kvecs[nk1 - 1, :]
            for nk2 in range(1, so._nkpts + 1):  # spin orbit
                kso = so._kvecs[nk2 - 1, :]
                if (
                    self.isclose(kso[0], knoso[0])
                    and self.isclose(kso[1], knoso[1])
                    and self.isclose(kso[2], knoso[2])
                ):  # do kpoints match?
                    if nk1 == nk2:
                        for c, e in enumerate(noso._occs[0, nk1 - 1, :]):
                            if e < 0.5:
                                cup = c
                                break
                        for c, e in enumerate(noso._occs[1, nk1 - 1, :]):
                            if e < 0.5:
                                cdn = c
                                break

                        nelec_list.append([cup, cdn, cup + cdn])
                        # total number of occupied bands at k point nk1

        n_arr = np.array(nelec_list)

        n_up = int(round(np.mean(n_arr[:, 0])))
        n_dn = int(round(np.mean(n_arr[:, 1])))
        n_tot = int(round(np.mean(n_arr[:, 2])))

        nelec = int(n_tot)

        # noso_homo_up = np.max(noso_bands[0, :, n_up - 1])
        # noso_lumo_up = np.min(noso_bands[0, :, n_up])
        # noso_homo_dn = np.max(noso_bands[1, :, n_dn - 1])
        # noso_lumo_dn = np.min(noso_bands[1, :, n_dn])

        so_homo = np.max(so_bands[0, :, nelec - 1])
        so_lumo = np.min(so_bands[0, :, nelec])

        # noso_direct_up = np.min(noso_bands[0, :, n_up] -
        # noso_bands[0, :, n_up - 1])
        # noso_direct_dn = np.min(noso_bands[1, :, n_dn] -
        # noso_bands[1, :, n_dn - 1])

        so_direct = np.min(so_bands[0, :, nelec] - so_bands[0, :, nelec - 1])

        noso_direct = 1000000.0
        noso_homo = -10000000.0
        noso_lumo = 100000000.0
        for i in range(noso_bands.shape[1]):
            homo_k = max(
                noso_bands[0, i, n_up - 1], noso_bands[1, i, n_dn - 1]
            )
            lumo_k = min(noso_bands[0, i, n_up], noso_bands[1, i, n_dn])
            noso_direct = min(noso_direct, lumo_k - homo_k)

            noso_homo = max(noso_homo, homo_k)
            noso_lumo = min(noso_lumo, lumo_k)

        gamma_k = []
        kpoints = []
        np.set_printoptions(precision=4)

        x = []
        y = []

        nelec_tot = 0.0
        for nk1 in range(1, noso._nkpts + 1):  # no spin orbit kpoints loop
            knoso = noso._kvecs[nk1 - 1, :]
            for nk2 in range(1, so._nkpts + 1):  # spin orbit
                kso = so._kvecs[nk2 - 1, :]
                if (
                    self.isclose(kso[0], knoso[0])
                    and self.isclose(kso[1], knoso[1])
                    and self.isclose(kso[2], knoso[2])
                ):  # do kpoints match?

                    # changes section 2
                    if nk1 == nk2:
                        nelec_up = n_arr[nk1 - 1, 0]
                        nelec_dn = n_arr[nk1 - 1, 1]
                        nelec_tot = n_arr[nk1 - 1, 2]

                        gamma_k.append(nelec_tot)
                        kpoints.append(kso)
                        Mmn = 0.0
                        vnoso = noso.readBandCoeff(
                            ispin=1, ikpt=nk1, iband=1, norm=False
                        )
                        # Getting size of matrices
                        n_noso1 = vnoso.shape[0]
                        # number of plane waves in noso
                        vnoso = noso.readBandCoeff(
                            ispin=2, ikpt=nk1, iband=1, norm=False
                        )
                        # spin, kpt, bands, number of plane wave
                        # n_noso2 = vnoso.shape[0]
                        vso = so.readBandCoeff(
                            ispin=1, ikpt=nk2, iband=1, norm=False
                        )
                        n_so = vso.shape[0]  # number of plane waves in so
                        vs = min(n_noso1 * 2, n_so)
                        # Vnono,and so holds wavefunctions
                        Vnoso = np.zeros((vs, nelec_tot), dtype=complex)
                        Vso = np.zeros((vs, nelec_tot), dtype=complex)

                        # prepare matricies,
                        # putting the wavefunction coeffients
                        # align so that they have same structure as SOC case
                        # which has both spin and down together

                        for n1 in range(1, nelec_up + 1):
                            Vnoso[0 : vs // 2, n1 - 1] = noso.readBandCoeff(
                                ispin=1, ikpt=nk1, iband=n1, norm=False
                            )[0 : vs // 2]
                        for n1 in range(1, nelec_dn + 1):
                            Vnoso[
                                vs // 2 : vs, n1 - 1 + nelec_up
                            ] = noso.readBandCoeff(
                                ispin=2, ikpt=nk1, iband=n1, norm=False
                            )[
                                0 : vs // 2
                            ]

                        for n1 in range(1, nelec_tot + 1):
                            t = so.readBandCoeff(
                                ispin=1, ikpt=nk2, iband=n1, norm=False
                            )
                            Vso[0 : vs // 2, n1 - 1] = t[0 : vs // 2]
                            Vso[vs // 2 : vs, n1 - 1] = t[
                                n_so // 2 : n_so // 2 + vs // 2
                            ]
                        # make orthonormal basis?
                        # Occupied bands, both up and down
                        # Generally to get wavefunctions, we have
                        # <psi><S><psi*>
                        # But as we do't have S from DFT saved
                        # we orthogalize the wavefunctions hoping
                        # they span the same space
                        Qnoso, num_noso = self.orth(Vnoso)

                        Qso, num_so = self.orth(Vso)

                        # evalute main expression

                        a = []
                        for n1 in range(0, nelec_tot):  # noso occupied bands
                            v1 = Qnoso[:, n1]  # number of plane waves, bands
                            aa = 0.0
                            # represents the non band-inverted electrons
                            for n2 in range(0, nelec_tot):  # so occupied bands
                                v2 = Qso[:, n2]
                                # Inner product of a nonsoc
                                # and soc wavefunction
                                # index n1 and n2

                                # The inner product
                                # of bands with and without SOC
                                # will be 1 for trivial bands

                                # Delta function of n1 and n2 if SOC is weak

                                t = np.dot(np.conj(v1), v2)
                                Mmn += t * t.conj()
                                aa += (t * t.conj()).real
                            a.append(aa)
                        gamma_k[-1] -= Mmn  # eq 4 in prb 90 125133
                        x.append(kso)
                        y.append(np.real(gamma_k[-1]))
                        if gamma_k[-1] > 0.5:
                            print(
                                "nk1 nk2 kpoint gamma_k ",
                                nk1,
                                nk2,
                                kso,
                                knoso,
                                np.real(gamma_k[-1]),
                                "!!!!!!!!!!",
                            )

        gmax = max(np.real(gamma_k))
        nkmax = np.argmax(np.real(gamma_k))
        kmax = kpoints[nkmax]

        print("------------------------------------")
        print()
        print("                   INDIRECT DIRECT      HOMO/LUMO (eV)")
        print(
            "no spin-orbit gaps",
            "{:+.3f}".format(float(noso_lumo - noso_homo)),
            "{:+.3f}".format(noso_direct),
            "   ",
            [noso_homo, noso_lumo],
        )
        print(
            "spin-orbit gaps   ",
            "{:+.3f}".format(float(so_lumo - so_homo)),
            "{:+.3f}".format(so_direct),
            "   ",
            [so_homo, so_lumo],
        )
        print()
        info = {}
        print("gamma max", np.real(gmax), " at k =  ", kmax)
        info["spillage"] = np.real(gmax)
        info["spillage_k"] = np.real(gamma_k).tolist()
        info["kpoints"] = [kk.tolist() for kk in kpoints]
        info["kmax"] = kmax.tolist()
        info["noso_direct"] = noso_direct
        info["so_direct"] = so_direct
        info["x_axis"] = [xx.tolist() for xx in x]
        info["y_axis"] = [yy.tolist() for yy in y]
        info["so_lumo"] = so_lumo
        info["so_homo"] = so_homo
        info["noso_lumo"] = noso_lumo
        info["noso_homo"] = noso_homo
        return info


# wf_so='/rk2/knc6/JARVIS-DFT/Bulk9-at30/mp-22260_PBEBO/MAIN-SOCSCFBAND-JVASP-59757_mp-22260/WAVECAR'
# wf_noso='/rk2/knc6/JARVIS-DFT/Bulk9-at30/mp-22260_PBEBO/MAIN-MAGSCFBAND-JVASP-59757_mp-22260/WAVECAR'

# spl = Spillage(wf_noso=wf_noso, wf_so=wf_so)
# info = spl.overlap_so_spinpol()
"""
if __name__ == "__main__":
    # JVASP-1044
    wf_noso = "../../tests/testfiles/analysis/spillage/WAVECAR.nosoc"
    wf_so = "../../tests/testfiles/analysis/spillage/WAVECAR.soc"
    spl = Spillage(wf_noso=wf_noso, wf_so=wf_so)
    info = spl.overlap_so_spinpol()
    print(info["spillage"])
"""
